/**
*
* $Id: X20.cpp
* (c) 2004 Sputnik -- erik_h_anderson@hotmail.com
* Based on CVEL Centaur-G code
* Ervins Reinverts -- ervins@gmail.com
* Based on CVEL Fregat code
* (c) 2004 David Henderson -- orbiter@aibs.org.uk
* Licence LGPL -- http://www.gnu.org/licenses/lgpl.txt
*
* You may use this source code in your addon provided that if you
* publish your addon either privately or publicly the end user
* recieves a copy of the source code including any modifications
* you may have made. 
*
*/


#include "x20.h"



// 1. vertical lift component (wings and body)
// Borrowed from Mike Norman's Shuttle code

void VLiftCoeff (double aoa, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*15.0;
	static const double istep = 1.0/step;
	static const int nabsc = 25;
	//                              -180 -165  -150 -135  -120 -105 -90 -75.0  -60.0  -45.0  -30.0  -15.0   0.0   15.0      30.0     45.0  60.0  75.0  90.0 105.0  120.0  135.0  150.0  165.0 180.0  
	static const double CL[nabsc] = {0.1, 0.17, 0.2, 0.2, 0.17, 0.1, 0, -0.11, -0.24, -0.38,  -0.5,  -0.5,  0.02, 0.6955,    0.66,   0.42, 0.24, 0.10, 0.0, -0.16, -0.26, -0.29, -0.24, -0.1, 0.1};
	static const double CM[nabsc] = {  0,    0,   0,   0,    0,   0, 0,     0,     0,  0.002,  0.004, 0.0025,0.0012,      0,-0.0012,-0.0007,    0,    0,   0,     0,     0,     0,     0,    0,   0};
	// lift and moment coefficients from -180 to 180 in 15 degree steps.
	// This uses a documented lift slope of 0.0437/deg, everything else is rather ad-hoc

	// change for drag increase at high Mach:
	// Subsonic Equivalents

	double Cl0, Cl1, Cd0, Cd1, Cd2;
	if(M>=5.0) {
		Cl0 = -0.187040;
		Cl1 =  1.675600;
		Cd0 =  0.068540;
		Cd1 = -0.412900;
		Cd2 =  1.539960;
		*cl = Cl0 + Cl1*aoa;
		*cd = Cd0 + Cd1*aoa + Cd2*aoa* aoa;
	}
 	aoa += PI;
	int idx = max (0, min (23, (int)(aoa*istep)));
	double d = aoa*istep - idx;
	*cm = (CM[idx] + (CM[idx+1]-CM[idx])*d);
	if(M<1.25) { 
 		*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
 		*cd = 0.04 + oapiGetInducedDrag (*cl, 2.1, 0.82);
	}
	else if(M<5.0) { 
 		*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
 		*cd = 0.07 + oapiGetInducedDrag (*cl, 2.1, 0.65);
// 		*cd = 0.06 + oapiGetInducedDrag (*cl, 2.266, 0.6);
	}
}

// 2. horizontal lift component (vertical stabilisers and body)

void HLiftCoeff (double beta, double M, double Re, double *cl, double *cm, double *cd)
{
	const int nabsc = 8;
	int i;
	static const double BETA[nabsc] = {-180*RAD,-135*RAD,-90*RAD,-45*RAD,45*RAD,90*RAD,135*RAD,180*RAD};
	static const double CL[nabsc]   = {       0,    +0.3,      0,   -0.3,  +0.3,     0,   -0.3,      0};
	for (i = 0; i < nabsc-1 && BETA[i+1] < beta; i++);
	*cl = CL[i] + (CL[i+1]-CL[i]) * (beta-BETA[i]) / (BETA[i+1]-BETA[i]);
	*cm = 0.0;
	*cd = 0.015 + oapiGetInducedDrag (*cl, 1.5, 0.6) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
	
	if (M > 1.0) *cd *= 2.2;
	else *cd *= 0.6;
}


X20::X20 (OBJHANDLE hObj, int fmodel)
: COMMON_VESSEL_EX (hObj)
{
	hX20				= oapiLoadMeshGlobal ("X20/X-20");
	hX20Bay1			= oapiLoadMeshGlobal ("X20/X20Bay1");
	hX20Bay2			= oapiLoadMeshGlobal ("X20/X20Bay2");
	hTopHatch			= oapiLoadMeshGlobal ("X20/TopHatch");	
	hEntryShield		= oapiLoadMeshGlobal ("X20/EntryShield");
	hSeat				= oapiLoadMeshGlobal ("X20/MOOSE");
	hAstro				= oapiLoadMeshGlobal ("gem-eva1");
	hTitanTransition	= oapiLoadMeshGlobal ("X20/TitanTransition");
	hTranstage			= oapiLoadMeshGlobal ("X20/X20Transtage");
	hCentaur			= oapiLoadMeshGlobal ("X20/X20Centaur");
	hChariot			= oapiLoadMeshGlobal ("X20/X20Chariot");
	hDummy1				= oapiLoadMeshGlobal ("X20/dummy1");
	hDummy2				= oapiLoadMeshGlobal ("X20/dummy2");
	hDummy3				= oapiLoadMeshGlobal ("X20/dummy3");
	hDummy4				= oapiLoadMeshGlobal ("X20/dummy4");
	hPit				= oapiLoadMeshGlobal ("X20/X-20_Pit");

	tex = oapiRegisterExhaustTexture("RealExhaust");

	/* set the propellant handles to NULL so I can check if I need to 
	/* create them in the SetMainStage() function. */
	ph_first=NULL; 
	ph_x20=NULL;
	ph_deorbit=NULL;
	bShieldJettisoned=false;
	bDeorbitFired=false;
	bThrustTerminated=false;
	bEjected=false;
	bEVA=false;
	gear_status=DOOR_CLOSED;
	gear_proc=0.0;
	pldoors_status=DOOR_CLOSED;
	pldoors_proc=0.0;
	brakes_status=DOOR_CLOSED;
	brakes_proc=0.0;
	hatch_status=DOOR_CLOSED;
	hatch_proc=0.0;
	oldsimt = -0.01;
	Vold = _V(0,0,0);
	bAbort = false;
	AbortChecked = false;
	atchmass = 0.0;
	AutoPilot = false;
	stagetype = 1;
	Pitch_Cntrl_Sig = 0.0;

	mfds_left.ngroup  = 0;
	mfds_right.ngroup = 202;
	mfds_left.nmesh   = 0;
	mfds_right.nmesh  = 0;
	huds.ngroup       = 0;
	huds.nmesh        = 6;
	huds.hudcnt       = CAMERA_OFS + _V(0,0.1,1.5);
	huds.size         = 0.252;


	key = (char*)VirtualAlloc(NULL,256,MEM_COMMIT,PAGE_READWRITE);
	message = (char*)VirtualAlloc(NULL,256,MEM_COMMIT,PAGE_READWRITE);

	//		     			   child       pos           dir       rot      handle
	Baypos = CreateAttachment (false, _V(0,0,4.8),   _V(0,0,1), _V(0,-1,0), "X");

	CreateAttachment		(true, _V(0,1.2,4.8),   _V(0,-1,0), _V(1,0,0), "R1");

	Dock0 = CreateDock(_V(0,0.9,4.8),_V(0,1,0),_V(0,0,1));

	char classname[30];
	int i;
	strncpy( classname, GetClassName(), 29 );
	classname[29]=0;
	for( i = 0 ; i < strlen( classname ) ; i++ ) {
		classname[i] = toupper(classname[i]);
	}
	if( strstr( classname, "X20WTRANSTAGE" ) )
		stage = 0;
	else if( strstr( classname, "X20WCENTAUR" ) ) {
		stage = 0;
		stagetype = 2;
	}
	else if( strstr( classname, "X20WCHARIOT" ) ) {
		stage = 0;
		stagetype = 3;
	}
	else if( strstr( classname, "X20WDEORBIT" ) )
		stage = 1;
	else if( strstr( classname, "X20NOSHIELD" ) ) {
		stage = 2;
		bShieldJettisoned=true;
	}
	else
		stage = 2;

	MyID=ConnectToOrbiterSoundDLL3(GetHandle());
//		havesound=false;
//	else
//	{
		havesound=true;
		RequestLoadVesselWave3(MyID,PLAYGEAR,"Sound\\X20\\gear.wav",BOTHVIEW_FADED_FAR);
		RequestLoadVesselWave3(MyID,PLAYBANG,"Sound\\X20\\stg2sep.wav",BOTHVIEW_FADED_FAR);
		RequestLoadVesselWave3(MyID,PLAYJETT,"Sound\\X20\\fasep.wav",BOTHVIEW_FADED_FAR);
//	}

	anim_gear = 0.0;
	anim_pldoors = 0.0;
	anim_pldoors2 = 0.0;
	anim_brakes = 0.0;
	anim_hatch = 0.0;

	/* CVE-lite function ! */
	/* IMPORTANT */
	/* This function loops through the pending GDPS requests. Have it run
	/* in the constructor so it can set up things like the InitFuelLevel
	/* (initial fuel level) variable, and load the payloads */
	CheckCentralMessage();   

	static UINT LGearDoorGrp[1] = {17};
	static MGROUP_ROTATE LGearDoor (7, LGearDoorGrp, 1,
		_V(-2.0,-1.26,-2.64), _V(0,0,1), (float)(-90*RAD));
	static UINT LGearGrp[1] = {5};
	static MGROUP_TRANSLATE LGear (7, LGearGrp, 1,
		_V( 0.0,-0.71,0.0));
	static UINT LSuptGrp[1] = {7};
	static MGROUP_ROTATE LSupt (7, LSuptGrp, 1,
		_V(0.0,-1.1, 2.2), _V(1,0,0), (float)( 70*RAD));

	static UINT RGearDoorGrp[1] = {16};
	static MGROUP_ROTATE RGearDoor (7, RGearDoorGrp, 1,
		_V( 2.0,-1.26,-2.64), _V(0,0,1), (float)(90*RAD));
	static UINT RGearGrp[1] = {6};
	static MGROUP_TRANSLATE RGear (7, RGearGrp, 1,
		_V( 0.0,-0.71,0.0));
	static UINT RSuptGrp[1] = {8};
	static MGROUP_ROTATE RSupt (7, RSuptGrp, 1,
		_V( 0.0,-1.1, 2.2), _V(1,0,0), (float)( 70*RAD));

	static UINT NoseGearDoorGrp[1] = {18};
	static MGROUP_ROTATE NoseGearDoor (7, NoseGearDoorGrp, 1,
		_V(-0.20,-0.70,11.0), _V(0,0.18,0.85), (float)(-90*RAD));
	static UINT NoseGearGrp[1] = {4};
	static MGROUP_TRANSLATE NoseGear (7, NoseGearGrp, 1,
		_V( 0.0,-0.60,0.0));
	static UINT NoseSuptGrp[1] = {0};
	static MGROUP_ROTATE NoseSupt (7, NoseSuptGrp, 1,
		_V( 0.0,-1.1, 9.0), _V(1,0,0), (float)( 80*RAD));

// Known Groups:

// 0-10 silvery
// 11 Transtage
// 12-20 Body

// 0 Nose Supt
// 1 R Gear Box
// 2 L Gear Box
// 3, 4  Nose Gear Box
// 5, 6  L Skid, R Skid
// 12 R Rudder
// 13 L Rudder
// 14 R Elevon
// 15 L Elevon


// 11 PLbox

// Bad Groups:
// 0	L Rudder
// 1	R Rudder
// 2	R Elevon
// 3	L Elevon
// 4	Rt Gear Door
// 5	Lt gear door
// 6    Nose gear door
// 9 10 Main Body
// 11	Rt gear box
// 12	Nose gear box
// 13   Nose skid
// 14   Rt Skid
// 15   Lt Skid
// 16 17 Gear Struts  
// 17   now Lt Skid
// 18   Q-ball
// 19   Cockpitbox  now Q-ball
// 20   Nose Strut  now Cockpitbox
// 18 19 20  Q-ball, plbox, cockpitbox


	static UINT LPLDoorGrp[1] = {0};
	static MGROUP_ROTATE LPLDoor (2, LPLDoorGrp, 1,
		_V(-0.92,0.13,5.1), _V(0,0,1), (float)( 130*RAD));
	static UINT RPLDoorGrp[1] = {0};
	static MGROUP_ROTATE RPLDoor (1, RPLDoorGrp, 1,
		_V( 0.92,0.13,5.1), _V(0,0,1), (float)(-130*RAD));
	static UINT HatchGrp[1] = {0};
	static MGROUP_ROTATE Hatch (4, HatchGrp, 1,
		_V( 0.0,0.75,6.78), _V(1,0,0), (float)( 155*RAD));
//		_V(-0.422,0.8,6.25), _V(0,0,1), (float)( 130*RAD));

	anim_gear = CreateAnimation (0);
	anim_pldoors = CreateAnimation(0);
//	anim_pldoors2 = CreateAnimation(0);
	anim_hatch = CreateAnimation(0);
	anim_yaw = CreateAnimation(0);
	anim_pitch = CreateAnimation(0.5);
	anim_roll = CreateAnimation(0.5);

	AddAnimationComponent (anim_gear, 0.05, 0.4, &LGearDoor);
	AddAnimationComponent (anim_gear, 0.4, 0.9, &LGear);
	AddAnimationComponent (anim_gear, 0.4, 0.95, &LSupt);
	AddAnimationComponent (anim_gear, 0.05, 0.4, &RGearDoor);
	AddAnimationComponent (anim_gear, 0.4, 0.9, &RGear);
	AddAnimationComponent (anim_gear, 0.4, 0.95, &RSupt);
	AddAnimationComponent (anim_gear, 0.4, 0.9, &NoseGear);
	AddAnimationComponent (anim_gear, 0.1, 0.4, &NoseGearDoor);
	AddAnimationComponent (anim_gear, 0.5, 0.9, &NoseSupt);

	AddAnimationComponent (anim_pldoors, 0.1, 0.9, &LPLDoor);
	AddAnimationComponent (anim_pldoors, 0.1, 0.9, &RPLDoor);
	AddAnimationComponent (anim_hatch, 0.1, 0.9, &Hatch);

//	MGROUP_TRANSFORM *ADIYaw;
//	MGROUP_TRANSFORM *ADIPitch;

/*
   static UINT BAGrp0[1] = {67};
   
   static MGROUP_ROTATE bank2 (1,BAGrp0, 1, _V(-4.845047E-02,2.742418E-02,6.320879E-02), _V(0,0,1), (float)(-360*RAD));
   
   
   static UINT PAGrp1[1] = {56};
   
   pitch1 = new MGROUP_ROTATE (1,PAGrp1, 1, _V(-4.845047E-02,2.742418E-02,6.320879E-02), _V(1,0,0), (float)(-360*RAD));
   

   
   anim_pitch = CreateAnimation (0.5);
   anim_bank = CreateAnimation (0.5);
   
   parent1 = AddAnimationComponent  (anim_bank, 0, 1, &bank2);
   AddAnimationComponent  (anim_pitch, 0, 1, pitch1, parent1);
   
 */
	
//	static MGROUP_ROTATE ADIYaw (LOCALVERTEXLIST, MAKEGROUPARRAY(&yawvec), 1, _V(0.0,-0.232,7.463), _V(0,1,0), (float)( 360*RAD));
//	static MGROUP_ROTATE ADIPitch  (LOCALVERTEXLIST, MAKEGROUPARRAY(&pitchvec), 1, _V(0.0,-0.232,7.463), _V(-1,0,0), (float)( 180*RAD));
//	ADIYaw = new MGROUP_ROTATE (LOCALVERTEXLIST, MAKEGROUPARRAY(&yawvec), 1, _V(0.0,-0.232,7.463), _V(0,1,0), (float)( 360*RAD));
//	ADIPitch = new MGROUP_ROTATE (LOCALVERTEXLIST, MAKEGROUPARRAY(&pitchvec), 1, _V(0.0,-0.232,7.463), _V(-1,0,0), (float)( 180*RAD));

	static UINT ADIYawGrp[1] = {0};
	MGROUP_TRANSFORM *ADIYaw = new MGROUP_ROTATE (0, ADIYawGrp, 1,
		_V(0.0,-0.232,7.463), _V(0,1,0), (float)( 360*RAD));
	static UINT ADIPitchGrp[1] = {0};
	MGROUP_TRANSFORM *ADIPitch = new MGROUP_ROTATE (0, ADIPitchGrp, 1,
		_V(0.0,-0.232,7.463), _V(-1,0,0), (float)( 180*RAD));

	static UINT ADIRollGrp[1] = {0};
	MGROUP_TRANSFORM *ADIRoll = new MGROUP_ROTATE (0, ADIRollGrp, 1,
		_V(0.0,-0.232,7.463), _V(0,0,-1), (float)( 360*RAD));


	animh_yaw = AddAnimationComponent (anim_yaw, 0, 1, ADIYaw);
	animh_pitch = AddAnimationComponent (anim_pitch, 0, 1, ADIPitch);
	animh_roll = AddAnimationComponent (anim_roll, 0, 1, ADIRoll);

	// ***** Elevator animation *****
	static UINT LElevatorGrp[1] = {14};
	static MGROUP_ROTATE LElevator (7, LElevatorGrp, 1,
		_V(-4.2,-1.0,-0.42), _V( 1,0,0), (float)( 25*RAD));
	anim_lelevator = CreateAnimation (0.5);
	AddAnimationComponent (anim_lelevator, 0, 1, &LElevator);
	static UINT RElevatorGrp[1] = {15};
	static MGROUP_ROTATE RElevator (7, RElevatorGrp, 1,
		_V( 4.2,-1.0,-0.42), _V(-1,0,0), (float)(-25*RAD));
	anim_relevator = CreateAnimation (0.5);
	AddAnimationComponent (anim_relevator, 0, 1, &RElevator);

	// ***** Elevator trim animation *****
	static UINT LElevatorTrimGrp[1] = {14};
	static MGROUP_ROTATE LElevatorTrim (7, LElevatorTrimGrp, 1,
		_V(-4.2,-1.0,-0.42),  _V( 1,0,0), (float)( 10*RAD));
	anim_lelevatortrim = CreateAnimation (0.5);
	AddAnimationComponent (anim_lelevatortrim, 0, 1, &LElevatorTrim);
	static UINT RElevatorTrimGrp[1] = {15};
	static MGROUP_ROTATE RElevatorTrim (7, RElevatorTrimGrp, 1,
		_V( 4.2,-1.0,-0.42),  _V( 1,0,0), (float)( 10*RAD));
	anim_relevatortrim = CreateAnimation (0.5);
	AddAnimationComponent (anim_relevatortrim, 0, 1, &RElevatorTrim);

	// ***** Aileron animation *****
	static UINT LAileronGrp[1] = {14};
	static MGROUP_ROTATE LAileron (7, LAileronGrp, 1,
		_V(-4.2,-1.0,-0.42),  _V( 1,0,0), (float)( 20*RAD));
	anim_laileron = CreateAnimation (0.5);
	AddAnimationComponent (anim_laileron, 0, 1, &LAileron);

	static UINT RAileronGrp[1] = {15};
	static MGROUP_ROTATE RAileron (7, RAileronGrp, 1,
		_V( 4.2,-1.0,-0.42),  _V( 1,0,0), (float)(-20*RAD));
	anim_raileron = CreateAnimation (0.5);
	AddAnimationComponent (anim_raileron, 0, 1, &RAileron);

	// ***** Rudder animation *****
	static UINT RRudderGrp[1] = {13};
	static MGROUP_ROTATE RRudder (7, RRudderGrp, 1,
		_V( 4.4,0,-2.5), _V(0,1,0), (float)(-30*RAD));
	static UINT LRudderGrp[1] = {12};
	static MGROUP_ROTATE LRudder (7, LRudderGrp, 1,
		_V(-4.4,0,-2.5), _V(0,1,0), (float)(-30*RAD));
	anim_lrudder = CreateAnimation (0.5);
	anim_rrudder = CreateAnimation (0.5);
	AddAnimationComponent (anim_rrudder, 0, 1, &RRudder);
	AddAnimationComponent (anim_lrudder, 0, 1, &LRudder);

		// ***** Speedbrake animation *****
//	static UINT RRudderGrp[1] = {22};
	static MGROUP_ROTATE RRudder2 (7, RRudderGrp, 1,
		_V( 4.4,0,-2.5), _V(0,1,0), (float)(-20*RAD));
//	static UINT LRudderGrp[1] = {24};
	static MGROUP_ROTATE LRudder2 (7, LRudderGrp, 1,
		_V(-4.4,0,-2.5), _V(0,1,0), (float)( 20*RAD));
	anim_brakes = CreateAnimation (0);
	AddAnimationComponent (anim_brakes, 0, 1, &RRudder2);
	AddAnimationComponent (anim_brakes, 0, 1, &LRudder2);

	CreateVariableDragElement (&brakes_proc, 0.7, _V(0, 0, 0));

	CreateAirfoil (LIFT_VERTICAL, _V(0,0,-0.2), VLiftCoeff, 7, 32, 2.85);
	// wing and body lift+drag components

	CreateAirfoil (LIFT_HORIZONTAL, _V(0,0,-5.5), HLiftCoeff, 8, 8, 0.625);
	// vertical stabiliser and body lift and drag components

	CreateControlSurface (AIRCTRL_ELEVATOR,     4.5,  0.7, _V(  4.2,0,-5.67), AIRCTRL_AXIS_XPOS, anim_lelevator);
	CreateControlSurface (AIRCTRL_ELEVATOR,     4.5,  0.7, _V( -4.2,0,-5.67), AIRCTRL_AXIS_XPOS, anim_relevator);
	CreateControlSurface (AIRCTRL_RUDDER,       0.3,  0.7, _V(  9.5,0,-5.75), AIRCTRL_AXIS_YPOS, anim_lrudder);
	CreateControlSurface (AIRCTRL_RUDDER,       0.3,  0.7, _V( -9.5,0,-5.75), AIRCTRL_AXIS_YPOS, anim_rrudder);
	CreateControlSurface (AIRCTRL_AILERON,      1.8, -0.65, _V(-14.8,0,-5.58), AIRCTRL_AXIS_XPOS, anim_raileron);
	CreateControlSurface (AIRCTRL_AILERON,      1.8, -0.65, _V( 14.8,0,-5.58), AIRCTRL_AXIS_XNEG, anim_laileron);
	CreateControlSurface (AIRCTRL_ELEVATORTRIM, 1.8,  0.7, _V(  4.2,0,-5.67), AIRCTRL_AXIS_XPOS, anim_lelevatortrim);
	CreateControlSurface (AIRCTRL_ELEVATORTRIM, 1.8,  0.7, _V( -4.2,0,-5.67), AIRCTRL_AXIS_XPOS, anim_relevatortrim);

	CreateVariableDragElement (&gear_proc, 0.8, _V(0, -1, 0));

	PARTICLESTREAMSPEC rps = {
		0, 20, 20, 0, 0.03, 0.5, 100, 3, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_FLAT, 1, 1, PARTICLESTREAMSPEC::ATM_PLIN, 6e7, 12e7
	};
	AddReentryStream (&rps);
	retex = oapiRegisterReentryTexture("reentry");
	SetReentryTexture(retex,6e6,1,1);


	Firstsweep = true;

}


X20::~X20()
{

	VirtualFree(key,256,MEM_DECOMMIT);
	VirtualFree(message,256,MEM_DECOMMIT);
}

void X20::UpdateAnimations()
{	

	// animate ADI ball

  pitch_proc=GetPitch();
  pitch_proc*=(1/(PI));
//  pitch_proc*=0.5;
  pitch_proc+=0.5;
  oapiGetHeading(GetHandle(),&yaw_proc);
  yaw_proc*=(1/(2*PI));
  roll_proc=GetBank();
  roll_proc*=(1/(2*PI));
  roll_proc+=0.5;
/*//		This isn't working right; temporarily remove it
    sprintf(oapiDebugString(),"yaw %4f pitch %4f roll %4f",yaw_proc,pitch_proc,roll_proc);

	SetAnimation (anim_yaw, yaw_proc);
	SetAnimation (anim_pitch, pitch_proc);
	SetAnimation (anim_roll, roll_proc);
*/

	// animate landing gear
	if (gear_status >= DOOR_CLOSING) {
		double da = oapiGetSimStep() * GEAR_OPERATING_SPEED;
		if (gear_status == DOOR_CLOSING) {
			if (gear_proc > 0.0)
				gear_proc = max (0.0, gear_proc-da);
			else {
				gear_status = DOOR_CLOSED;
			}
		} else  { // door opening
			if (gear_proc < 1.0)
				gear_proc = min (1.0, gear_proc+da);
			else {
				gear_status = DOOR_OPEN;
			}
		}
		SetAnimation (anim_gear, gear_proc);
	}

	// animate payload bay doors
	if (pldoors_status >= DOOR_CLOSING) {
		double da = oapiGetSimStep() * PLDOORS_OPERATING_SPEED;
		if (pldoors_status == DOOR_CLOSING) {
			if (pldoors_proc > 0.0)
				pldoors_proc = max (0.0, pldoors_proc-da);
			else {
				pldoors_status = DOOR_CLOSED;
			}
		} else  { // door opening
			if (pldoors_proc < 1.0)
				pldoors_proc = min (1.0, pldoors_proc+da);
			else {
				pldoors_status = DOOR_OPEN;
			}
		}
		SetAnimation (anim_pldoors, pldoors_proc);
//		SetAnimation (anim_pldoors2, pldoors_proc);
	}

	// animate speedbrakes
	if (brakes_status >= DOOR_CLOSING) {
		double da = oapiGetSimStep() * BRAKES_OPERATING_SPEED;
		if (brakes_status == DOOR_CLOSING) {
			if (brakes_proc > 0.0)
				brakes_proc = max (0.0, brakes_proc-da);
			else {
				brakes_status = DOOR_CLOSED;
			}
		} else  { // door opening
			if (brakes_proc < 1.0)
				brakes_proc = min (1.0, brakes_proc+da);
			else {
				brakes_status = DOOR_OPEN;
			}
		}
		SetAnimation (anim_brakes, brakes_proc);
	}

	// animate top hatch
	if (!bEjected)
		if (hatch_status >= DOOR_CLOSING) {
			double da = oapiGetSimStep() * HATCH_OPERATING_SPEED;
			if (hatch_status == DOOR_CLOSING) {
				if (hatch_proc > 0.0)
					hatch_proc = max (0.0, hatch_proc-da);
				else {
					hatch_status = DOOR_CLOSED;
				}
			} else  { // door opening
				if (hatch_proc < 1.0)
					hatch_proc = min (1.0, hatch_proc+da);
				else {
					hatch_status = DOOR_OPEN;
				}
			}
			SetAnimation (anim_hatch, hatch_proc);
		}

}

void X20::ActivateLandingGear (DoorStatus action)
{
	gear_status = action;
}

void X20::RevertLandingGear ()
{
	ActivateLandingGear (gear_status == DOOR_CLOSED || gear_status == DOOR_CLOSING ?
						 DOOR_OPENING : DOOR_CLOSING);
}

void X20::ActivatePayloadBayDoors (DoorStatus action)
{
	pldoors_status = action;
}

void X20::RevertPayloadBayDoors ()
{
	ActivatePayloadBayDoors (pldoors_status == DOOR_CLOSED || pldoors_status == DOOR_CLOSING ?
						 DOOR_OPENING : DOOR_CLOSING);
}

void X20::ActivateBrakes (DoorStatus action)
{
	brakes_status = action;
}

void X20::RevertBrakes ()
{
	ActivateBrakes (brakes_status == DOOR_CLOSED || brakes_status == DOOR_CLOSING ?
						 DOOR_OPENING : DOOR_CLOSING);
}

void X20::ActivateHatch (DoorStatus action)
{
	hatch_status = action;
}

void X20::RevertHatch ()
{
	ActivateHatch (hatch_status == DOOR_CLOSED || hatch_status == DOOR_CLOSING ?
						 DOOR_OPENING : DOOR_CLOSING);
}


void X20::SetGearParameters (double state)
{

	double geary = (-1*(gear_proc-1.0)*0.6);
	SetTouchdownPoints (_V(0,-1.8+geary,14.0), _V(-9.15,-2.0+geary,-3.0), _V(9.15,-2.0+geary,-3.0));
	if (state == 1.0 && !BrakesOn) {
		SetSurfaceFrictionCoeff (0.1, 0.4);
	} else {
		SetSurfaceFrictionCoeff (0.1, 0.4);		// Wire brushes; no brakes!
	}
}


void X20::JettisonEntryShield ()
{
	bShieldJettisoned = true;
	VESSELSTATUS vs;
	GetStatus (vs);
	vs.eng_main = vs.eng_hovr = 0.0;
	vs.fuel=1.0;
	vs.status=0;
	VECTOR3 ofs = ENTRYSHIELD_OFS;
	VECTOR3 vel = _V(0,20.0,+8.0);
	VECTOR3 rofs, rvel = {vs.rvel.x, vs.rvel.y, vs.rvel.z};
	Local2Rel (ofs, vs.rpos);
	GlobalRot (vel, rofs);
	vs.rvel.x = rvel.x+rofs.x;
	vs.rvel.y = rvel.y+rofs.y;
	vs.rvel.z = rvel.z+rofs.z;
	vs.vrot.x = 0.0;
	vs.vrot.y = 0.0;
	vs.vrot.z = 0.0;
	char name[256];
	strcpy (name, GetName()); strcat (name, "-Shield");
	VESSEL::Create (name, "X20/X20_Entryshield", vs);
	SetMainStage();
	PlayVesselWave3(MyID,PLAYJETT);
}

void X20::SeparateEVA ()
{
	VESSELSTATUS vs;
	GetStatus (vs);
	vs.eng_main = vs.eng_hovr = 0.0;
	vs.fuel=1.0;
	vs.status=0;
	VECTOR3 ofs = ASTRO_OFS;
	Local2Rel (ofs, vs.rpos);
	char name[256];
	strcpy (name, GetName()); strcat (name, "-PLT");
	oapiCreateVessel (name, "gemini_leva", vs);
}

void X20::TimeStep(  double simt )
{
	
	/* CVE-lite function ! */
	/* IMPORTANT */
	/* This function loops through the pending GPDS requests and returns a bool (true/false)
	/* if any GPDS requests have been implemented.
	/* If any have (eg adding a new payload), you will need to re-draw your ship, hence the
	/* call to the (repeatable) SetMainStage(); */

	ZeroMemory (key,256);
	ZeroMemory (message,256);


	if( CheckCentralMessage(key, message) ) {
		CheckCentralMessage();
		if (( strstr(key, "ON" ) ) && (stage == 0) && (!AutoPilot) && ( strstr(message, "AUTOPILOT" ) )) {
//			sprintf(oapiDebugString(),"Autopilot in continued operation");
			AutoPilot = true;
//			ActivateNavmode(NAVMODE_KILLROT);
			SetThrusterGroupLevel (thg_first,1.0);
		}
		while( CheckCentralMessage() ) ;
			SetMainStage();
	}

/*	erasecount++;						//  Every once in a while, check for spawned EVA's that are not focus-enabled,
	if (erasecount > 60) {				//  and delete them.
		erasecount = 0;
		char name[256];
			strcpy (name, GetName()); strcat (name, "-PLT");
			OBJHANDLE hvessel=oapiGetVesselByName(name);
			if (hvessel != 0) {
				VESSEL *remv = oapiGetVesselInterface(hvessel);
				if (remv->GetEnableFocus() == false)
					oapiDeleteVessel(hvessel);
			}
	}
*/
	if (AutoPilot) SetThrusterGroupLevel (thg_first,1.0);

	if (AutoPilot) {
		GetShipAirspeedVector(Rel_Vel);
		Velocity = sqrt((Rel_Vel.x*Rel_Vel.x) + (Rel_Vel.y*Rel_Vel.y) + (Rel_Vel.z*Rel_Vel.z));
		Altitude=GetAltitude();
		a = ((APO_ALT * 1000 + R_EARTH) + (Altitude + R_EARTH)) / 2;
		e = ((APO_ALT * 1000 + R_EARTH) / a ) - 1;
		h = sqrt(U_EARTH * a * ( 1 - (e * e)));
		V_periapsis = h / (Altitude + R_EARTH);
		MJD_Time = oapiGetSimMJD();
		GetElements( elements, MJD_Time );
		Alt_Apoapsis = (elements.a * ( 1 + elements.e )/1000) - 6371; 
		Alt_Periapsis = (elements.a * ( 1 - elements.e )/1000) - 6371; 
		
//		if (Velocity > V_periapsis) 
		if ((Alt_Periapsis > 120.0) || (Alt_Apoapsis > 200.0) || (Velocity > V_periapsis))
		{
//			sprintf(oapiDebugString(),"Autopilot Off");
			AutoPilot = false;
			ActivateNavmode(NAVMODE_KILLROT);
			SetAttitudeRotLevel( 0 , 0.0);
			SetAttitudeRotLevel( 1 , 0.0);
			SetAttitudeRotLevel( 2 , 0.0);
			SetThrusterGroupLevel (thg_first,0.0);
		}
		else {
			GetHorizonAirspeedVector(horiz);

			Pitch_SP = -0.03*horiz.y-0.85*GetAOA()*DEG;			

			if (Pitch_SP < 0.0)
				Pitch_SP = 0.0;
			if (Pitch_SP > 40.0)
				Pitch_SP = 40.0;

			Pitch_Error = GetPitch()*DEG - Pitch_SP;
			GetAngularVel( Angular_Velocity ); 
			Pitch_Error_Dot = Angular_Velocity.x * DEG; // get pitch change rate in degrees/s

			Pitch_Cntrl_Sig = (- 1.4 * Pitch_Error_Dot + 0.15 * Pitch_Error);

			if (Pitch_Cntrl_Sig > 1 )
			{
				Pitch_Cntrl_Sig = 1;
			}
			else if (Pitch_Cntrl_Sig < -1) 
			{
				Pitch_Cntrl_Sig = -1;
			}
			SetAttitudeRotLevel( 0 , Pitch_Cntrl_Sig);

			SetAttitudeRotLevel( 1, (DEG*GetSlipAngle() / 15.0)- (0.2*Angular_Velocity.y*DEG));
			Roll_Cntrl_Sig = 0.15*((GetBank()*180/PI)/-100.0) - (0.2*Angular_Velocity.z*DEG);
			SetAttitudeRotLevel( 2, Roll_Cntrl_Sig);

		}
//		else if ((e < 0.05) && (Rel_Vel.y > 100.0))
//			ActivateNavmode(NAVMODE_PROGRADE);
	}

// Check to see if the EVA guy is back inside.  If so, start drawing the pilot in his seat again
	if (bEVA) {
		char name[256];
		OBJHANDLE hvessel;
		strcpy (name, GetName()); strcat (name, "-PLT");
		hvessel=oapiGetVesselByName(name);
		if (hvessel == 0) {
			bEVA = false;
			SetMainStage();
		}
	}

	if ((stage==0) && (!AbortChecked)) {
		AbortChecked = true;
		GetShipAirspeedVector(Rel_Vel);
		Velocity = sqrt((Rel_Vel.x*Rel_Vel.x) + (Rel_Vel.y*Rel_Vel.y) + (Rel_Vel.z*Rel_Vel.z));
		if (Velocity < 4500.0)
			bAbort = true;
	}
	UpdateAnimations();
	SetGearParameters(gear_proc);
	if (bAbort) {
		if (stage==0) {
			JettisonPart();
			SetMainStage();
		}
		if (stage==1) SetThrusterGroupLevel (thg_first,1.0);
	}

	if ((stage==2) && (!bShieldJettisoned))
		if ((GetMachNumber() < 6.0) && (GetAtmPressure() > 100.0))
			JettisonEntryShield();
	if (stage==1) {
		if ((GetThrusterGroupLevel (thg_first) == 0.0) && (bDeorbitFired) && (!bThrustTerminated)) {
			bThrustTerminated = true;
			DelExhaust(Exh1);
			DelExhaust(Exh2);
			DelExhaust(Exh3);
			DelExhaust(Exh4);
//			DelExhaustStream(stream_a_1);
			DelExhaustStream(stream_b_1);
//			DelExhaustStream(stream_a_2);
			DelExhaustStream(stream_b_2);
//			DelExhaustStream(stream_a_3);
			DelExhaustStream(stream_b_3);
//			DelExhaustStream(stream_a_4);
			DelExhaustStream(stream_b_4);
			for (int i = 0; i < 4; i++) {
				SetThrusterIsp(th_first[i],ISP_DEORBIT_VAC/100);
				SetThrusterMax0(th_first[i],THRUST_DEORBIT_VAC/100);
			}
			PARTICLESTREAMSPEC exhaust_main = {
				0, 0.5, 1, 100, 0, 0.1, 50, 0.0, PARTICLESTREAMSPEC::EMISSIVE,
			PARTICLESTREAMSPEC::LVL_FLAT, 1, 1,
			PARTICLESTREAMSPEC::ATM_FLAT, 0.3, 0.3
			};
			exhaust_main.tex = oapiRegisterParticleTexture ("Contrail2");
			AddExhaustStream (th_first[0], _V(-0.271,-0.136,-2.8), &exhaust_main);
			AddExhaustStream (th_first[1], _V( 0.271,-0.136,-2.8), &exhaust_main);
			AddExhaustStream (th_first[2], _V(-0.271,-0.688,-2.8), &exhaust_main);
			AddExhaustStream (th_first[3], _V( 0.271,-0.688,-2.8), &exhaust_main);
			PlayVesselWave3(MyID,PLAYBANG);
		}
		if (GetThrusterGroupLevel (thg_first) > 0.05) 
			bDeorbitFired = true;
	}
	if ((bDeorbitFired) && (stage==1)) SetThrusterGroupLevel (thg_first,1.0);  // Act as a solid motor (burn to depletion)

	if (Firstsweep) {
		AttachSweep();
		if (stage==1)
			if ((GetPropellantMass(ph_deorbit) < 1.1) && (GetPropellantMass(ph_x20) < 1.1))
			{
			SetPropellantMass(ph_deorbit,DEORBIT_PROP_MASS);
			SetPropellantMass(ph_x20,RCS_PROP_MASS);
			}
		Firstsweep = false;
	}
/*
	dt = simt - oldsimt;
	GetShipAirspeedVector(Vnew);
	double lift = (Vnew.y - Vold.y)/dt;
	double drag = (Vnew.z - Vold.z)/dt;
	sprintf(oapiDebugString(),"L/D %1f",(lift/drag));
	Vold = Vnew;
	oldsimt = simt;
*/
}

/* CVE-lite function ! */
/* IMPORTANT */
/* VisualCreated() is a function passed through from CVE-lite as a direct replacement
/* for the DLLCLBK ovcVisualCreated() function.
/* You can't use ovcVisualCreated() with CVE-lite, so use this instead! */

void X20::VisualCreated ( VESSEL2 *vessel, VISHANDLE vis, int refcount) {
	/* local equivalent to the DLLCLBK ovcVisualCreated() */
	SetMainStage(); // this is a good place to ask for the initial setup.
}


void X20::SetClassCaps(  FILEHANDLE cfg )
{

	CheckCentralMessage();

	SetMainStage();
}

void X20::SaveState( FILEHANDLE scn )
{
	char cbuf[256];

	SaveDefaultState (scn);
	
	/* CVE-lite function ! */
	/* IMPORTANT */
	/* Call this after the normal SaveDefautState() to record attached payloads
	/* it the scenario file */


//	Undock(ALLDOCKS);

	oapiWriteScenario_int (scn, "CONFIGURATION", stage);
	oapiWriteScenario_int (scn, "SHIELDGONE", bShieldJettisoned);
	oapiWriteScenario_int (scn, "EJECTED", bEjected);
	oapiWriteScenario_int (scn, "EVA", bEVA);
	oapiWriteScenario_int (scn, "DEORBIT FIRED", bDeorbitFired);
	oapiWriteScenario_int (scn, "THRUST TERMINATED", bThrustTerminated);
	sprintf (cbuf, "%d %0.4f", gear_status, gear_proc);
	oapiWriteScenario_string (scn, "GEAR", cbuf);
	sprintf (cbuf, "%d %0.4f", pldoors_status, pldoors_proc);
	oapiWriteScenario_string (scn, "DOORS", cbuf);
	sprintf (cbuf, "%d %0.4f", brakes_status, brakes_proc);
	oapiWriteScenario_string (scn, "BRAKES", cbuf);
	sprintf (cbuf, "%d %0.4f", hatch_status, hatch_proc);
	oapiWriteScenario_string (scn, "HATCH", cbuf);

	/* save payload lines */
	SaveDefaultStateEx(scn);
}

void X20::LoadState(  FILEHANDLE scn, void *vs )
{
    char *line;

	while (oapiReadScenario_nextline (scn, line)) 
	{
		/* CVE-lite function ! */
		/* IMPORTANT */
		/* Call this for every line in the scenario file. It will automatically
		/* deduce which lines are for it, and load the payloads from the file. */

		/* load payload lines */
		if (!strnicmp (line, "CONFIGURATION", 13)) {
		    sscanf (line+13, "%d", &stage);
		} else if (!strnicmp (line, "SHIELDGONE", 10)) {
			sscanf (line+10, "%d", &bShieldJettisoned);
		} else if (!strnicmp (line, "EJECTED", 7)) {
			sscanf (line+7, "%d", &bEjected);
		} else if (!strnicmp (line, "EVA", 3)) {
			sscanf (line+3, "%d", &bEVA);
		} else if (!strnicmp (line, "DEORBIT FIRED", 12)) {
			sscanf (line+12, "%d", &bDeorbitFired);
		} else if (!strnicmp (line, "THRUST TERMINATED", 17)) {
			sscanf (line+17, "%d", &bThrustTerminated);
		} else if (!strnicmp (line, "GEAR", 4)) {
		sscanf (line+4, "%d%lf", &gear_status, &gear_proc);
		} else if (!strnicmp (line, "DOORS", 5)) {
		sscanf (line+5, "%d%lf", &pldoors_status, &pldoors_proc);
		} else if (!strnicmp (line, "BRAKES", 6)) {
		sscanf (line+6, "%d%lf", &brakes_status, &brakes_proc);
		} else if (!strnicmp (line, "HATCH", 5)) {
		sscanf (line+5, "%d%lf", &hatch_status, &hatch_proc);
		} else ParsePayloadString(line);

		ParseScenarioLineEx (line, vs);
	}

	/* always set the stage whenever something may have changed */
	SetMainStage();

	RevertLandingGear();
	RevertLandingGear();
	if (!bEjected) {
		RevertHatch();
		RevertHatch();
	}
	RevertBrakes();
	RevertBrakes();
	RevertPayloadBayDoors ();
	RevertPayloadBayDoors();

}


void X20::PlaceComponents()
{
	COMMON_MESH tmp;

		tmp.MeshHandle(hPit);
		tmp.MeshPosition(X20_OFS);
		AddMeshEx (tmp);

		tmp.MeshHandle(hX20Bay1);
		tmp.MeshPosition(BAY1_OFS);
		AddMeshEx (tmp);
		tmp.MeshHandle(hX20Bay2);
		tmp.MeshPosition(BAY2_OFS);
		AddMeshEx (tmp);
		if (!bEjected) {
			tmp.MeshHandle(hSeat);
			tmp.MeshPosition(EJSEAT_OFS);
			AddMeshEx (tmp);
			tmp.MeshHandle(hTopHatch);
			tmp.MeshPosition(TOPHATCH_OFS);
			AddMeshEx (tmp);
		}
		else {
			tmp.MeshHandle(hDummy1);
			tmp.MeshPosition(_V(0,0,0));
			AddMeshEx (tmp);
			tmp.MeshHandle(hDummy2);
			tmp.MeshPosition(_V(0,0,0));
			AddMeshEx (tmp);
		}

		if ((!bEVA) && (!bEjected)) {
			tmp.MeshHandle(hAstro);
			tmp.MeshPosition(ASTRO_OFS);
			AddMeshEx (tmp);
		}
		else {
			tmp.MeshHandle(hDummy3);
			tmp.MeshPosition(_V(0,0,0));
			AddMeshEx (tmp);
		}

		if (!bShieldJettisoned) {
			tmp.MeshHandle(hEntryShield);
			tmp.MeshPosition(ENTRYSHIELD_OFS);
			AddMeshEx (tmp);
		}
		else {
			tmp.MeshHandle(hDummy4);
			tmp.MeshPosition(_V(0,0,0));
			AddMeshEx (tmp);
		}
		tmp.MeshHandle(hX20);
		tmp.MeshPosition(X20_OFS);
		AddMeshEx (tmp);

	if (stage < 2) {
		tmp.MeshHandle(hTitanTransition);
		tmp.MeshPosition(TRANSITION_OFS);
		AddMeshEx (tmp);
	}
	if (stage == 0)  {
		if (stagetype == 1)
			tmp.MeshHandle(hTranstage);
		else if (stagetype == 2)
			tmp.MeshHandle(hCentaur);
		else if (stagetype == 3)
			tmp.MeshHandle(hChariot);
		tmp.MeshPosition(TRANSTAGE_OFS);
		AddMeshEx (tmp);
	}

}

/* IMPORTANT */
/* Always have a function that you can call as many times as you like that will re-draw
/* all you meshes, reset all your thrusters and check to make sure the propellent is
/* there. This should be called after any configuration change, or any GDPS line (see
/* TimeStep(), above) */

void X20::SetMainStage( )
{

	if (stage == 0) {
		if (stagetype == 1)
			SetTranstage();
		else if (stagetype == 2)
			SetCentaur();
		else if (stagetype == 3)
			SetChariot();
	}
	else if (stage == 1)
		SetDeorbitStage();
	else SetX20Stage();
}

void X20::SetTranstage( )
{


	/* CVE-lite function ! */
	/* ClearMeshesEx() will get rid of all the normal meshes, plus anything you've defined
	/* with the COMMON_MESH class (not used here) */

	/* erase all meshes and thrusters so we can start afresh */
	ClearMeshesEx();
	ClearThrusterDefinitions();

	PARTICLESTREAMSPEC contrail = 
	{
		0, 25.0, 7, 350, 0.3, 4.0, 7, 3.0, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};

	PARTICLESTREAMSPEC exhaust_main = {
		0, 3, 25, 350, 0.1, 0.2, 28, 1.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-5, 0.1
	};
	exhaust_main.tex = oapiRegisterParticleTexture ("Contrail2");
	SetSize (20);

	/* CVE-lite function ! */
	/* IMPORTANT */
	/* double CalculateMass() will return the all-up mass of the combined payloads and
	/* upper stages. Simply add this to your normal dry mass. */
	
	/* always add payload mass [CalculateMass()] to the vessel's empty mass */
	SetEmptyMass ( X20_DRYMASS + DEORBIT_DRYMASS + TRANSTAGE_DRYMASS + RCS_PROP_MASS + atchmass + CalculateMass() );


	SetTouchdownPoints (_V(0,-1.0,-3), _V(-.7,.7,-3), _V(.7,.7,-3));
	SetPMI (_V(18,19.5,3.25));
	SetCrossSections (_V(42.4,76.4,12.5));
	SetCW (0.1, 0.3, 1.4, 1.4);
	SetRotDrag (_V(0.7,0.7,1.2));
	SetPitchMomentScale (1e-6);
	SetBankMomentScale (1e-6);
//	SetLiftCoeffFunc (0); 
	SetTrimScale (0.05);
	SetCameraOffset(CAMERA_OFS);

	
	/* CVE-lite function ! */
	/* IMPORTANT */
	/* Draw() should be called to render all the loaded payloads, plus anything you've
	/* added with the COMMON_MESH class (not used here) */

	/* call draw to render the payloads, plus anything you've added from COMMON_MESH */
	PlaceComponents();
	Draw();
	SetMeshVisibilityMode(3,MESHVIS_ALWAYS);
	SetMeshVisibilityMode(0,MESHVIS_EXTERNAL|MESHVIS_VC);
	SetMeshVisibilityMode(6,MESHVIS_EXTERNAL|MESHVIS_VC);


	/* we're not using COMMON_MESH today, so draw our own */
//	AddMesh(hX20, &VECTOR_ZERO );
	

	/* CVE-lite variable ! */
	/* IMPORTANT */
	/* Use the keyword double InitFuelLevel multipled by the maximum mass to have the
	/* fuel levels automatically set from the GPDS payload string */

	/* verify that the propellent hasn't been created before making it */
	if (ph_first==NULL)  
		ph_first  = CreatePropellantResource(FIRST_PROP_MASS,FIRST_PROP_MASS*InitFuelLevel,1.0); 
	if (!ph_deorbit)  
		ph_deorbit  = CreatePropellantResource(DEORBIT_PROP_MASS,DEORBIT_PROP_MASS,1.0); 
//	if (!ph_x20)  
//		ph_x20  = CreatePropellantResource(RCS_PROP_MASS,RCS_PROP_MASS,1.0); 

	SetDefaultPropellantResource (ph_first); 

	th_first[0] = CreateThruster ( _V(0,0.72,-8.5), _V(0,0,1), THRUST_FIRST_VAC, ph_first, ISP_FIRST_VAC, ISP_FIRST_SL);
//	AddExhaustStream (th_first[0], _V(0.72,0,-7), &contrail);
//	AddExhaustStream (th_first[0], _V(0.72,0,-7), &exhaust_main);

	th_first[1] = CreateThruster ( _V(0,-0.72,-8.5), _V(0,0,1), THRUST_FIRST_VAC, ph_first, ISP_FIRST_VAC, ISP_FIRST_SL);
//	AddExhaustStream (th_first[1], _V(-0.72,0,-7), &contrail);
//	AddExhaustStream (th_first[1], _V(-0.72,0,-7), &exhaust_main);

	thg_first = CreateThrusterGroup (th_first, 2, THGROUP_MAIN);
	
	AddExhaust (th_first[0], 8.0, 0.9, tex);
	AddExhaust (th_first[1], 8.0, 0.9, tex);

	SetAttControlsTranstage();
	EnableTransponder(true);

}


void X20::SetCentaur( )
{


	/* CVE-lite function ! */
	/* ClearMeshesEx() will get rid of all the normal meshes, plus anything you've defined
	/* with the COMMON_MESH class (not used here) */

	/* erase all meshes and thrusters so we can start afresh */
	ClearMeshesEx();
	ClearThrusterDefinitions();
/*
	PARTICLESTREAMSPEC contrail = 
	{
		0, 25.0, 7, 350, 0.3, 4.0, 7, 3.0, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};

	PARTICLESTREAMSPEC exhaust_main = {
		0, 3, 25, 350, 0.1, 0.2, 28, 1.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-5, 0.1
	};
	exhaust_main.tex = oapiRegisterParticleTexture ("Contrail2");
*/
	SetSize (20);
	SURFHANDLE tex = oapiRegisterExhaustTexture("Exhaust_atsme");

	/* CVE-lite function ! */
	/* IMPORTANT */
	/* double CalculateMass() will return the all-up mass of the combined payloads and
	/* upper stages. Simply add this to your normal dry mass. */
	
	/* always add payload mass [CalculateMass()] to the vessel's empty mass */
	SetEmptyMass ( X20_DRYMASS + DEORBIT_DRYMASS + CENTAUR_DRYMASS + RCS_PROP_MASS + atchmass + CalculateMass() );


	SetTouchdownPoints (_V(0,-1.0,-3), _V(-.7,.7,-3), _V(.7,.7,-3));
	SetPMI (_V(18,19.5,3.25));
	SetCrossSections (_V(42.4,76.4,12.5));
	SetCW (0.1, 0.3, 1.4, 1.4);
	SetRotDrag (_V(0.7,0.7,1.2));
	SetPitchMomentScale (1e-6);
	SetBankMomentScale (1e-6);
//	SetLiftCoeffFunc (0); 
	SetTrimScale (0.05);
	SetCameraOffset(CAMERA_OFS);

	
	/* CVE-lite function ! */
	/* IMPORTANT */
	/* Draw() should be called to render all the loaded payloads, plus anything you've
	/* added with the COMMON_MESH class (not used here) */

	/* call draw to render the payloads, plus anything you've added from COMMON_MESH */
	PlaceComponents();
	Draw();
	SetMeshVisibilityMode(3,MESHVIS_ALWAYS);
	SetMeshVisibilityMode(0,MESHVIS_EXTERNAL|MESHVIS_VC);
	SetMeshVisibilityMode(6,MESHVIS_EXTERNAL|MESHVIS_VC);



	/* we're not using COMMON_MESH today, so draw our own */
//	AddMesh(hX20, &VECTOR_ZERO );
	

	/* CVE-lite variable ! */
	/* IMPORTANT */
	/* Use the keyword double InitFuelLevel multipled by the maximum mass to have the
	/* fuel levels automatically set from the GPDS payload string */

	/* verify that the propellent hasn't been created before making it */
	if (ph_first==NULL)  
		ph_first  = CreatePropellantResource(FIRST_PROP_MASS,FIRST_PROP_MASS*InitFuelLevel,1.0); 
	if (!ph_deorbit)  
		ph_deorbit  = CreatePropellantResource(DEORBIT_PROP_MASS,DEORBIT_PROP_MASS,1.0); 
//	if (!ph_x20)  
//		ph_x20  = CreatePropellantResource(RCS_PROP_MASS,RCS_PROP_MASS,1.0); 

	SetDefaultPropellantResource (ph_first); 

	th_first[0] = CreateThruster ( _V(0,0.72,-11.75), _V(0,0,1), THRUST_CENTAUR_VAC, ph_first, ISP_CENTAUR_VAC, ISP_CENTAUR_SL);
//	AddExhaustStream (th_first[0], _V(0.72,0,-7), &contrail);
//	AddExhaustStream (th_first[0], _V(0.72,0,-7), &exhaust_main);

	th_first[1] = CreateThruster ( _V(0,-0.72,-11.75), _V(0,0,1), THRUST_CENTAUR_VAC, ph_first, ISP_CENTAUR_VAC, ISP_CENTAUR_SL);
//	AddExhaustStream (th_first[1], _V(-0.72,0,-7), &contrail);
//	AddExhaustStream (th_first[1], _V(-0.72,0,-7), &exhaust_main);

	thg_first = CreateThrusterGroup (th_first, 2, THGROUP_MAIN);
	
	AddExhaust (th_first[0], 8.0, 0.9, tex);
	AddExhaust (th_first[1], 8.0, 0.9, tex);


	SetAttControlsTranstage();
	EnableTransponder(true);

}


void X20::SetChariot( )
{


	/* CVE-lite function ! */
	/* ClearMeshesEx() will get rid of all the normal meshes, plus anything you've defined
	/* with the COMMON_MESH class (not used here) */

	/* erase all meshes and thrusters so we can start afresh */
	ClearMeshesEx();
	ClearThrusterDefinitions();
	SURFHANDLE tex = oapiRegisterExhaustTexture ("FluorineExhaust");

	SetSize (20);

	/* CVE-lite function ! */
	/* IMPORTANT */
	/* double CalculateMass() will return the all-up mass of the combined payloads and
	/* upper stages. Simply add this to your normal dry mass. */
	
	/* always add payload mass [CalculateMass()] to the vessel's empty mass */
	SetEmptyMass ( X20_DRYMASS + DEORBIT_DRYMASS + CHARIOT_DRYMASS + RCS_PROP_MASS + atchmass + CalculateMass() );


	SetTouchdownPoints (_V(0,-1.0,-3), _V(-.7,.7,-3), _V(.7,.7,-3));
	SetPMI (_V(18,19.5,3.25));
	SetCrossSections (_V(42.4,76.4,12.5));
	SetCW (0.1, 0.3, 1.4, 1.4);
	SetRotDrag (_V(0.7,0.7,1.2));
	SetPitchMomentScale (1e-6);
	SetBankMomentScale (1e-6);
//	SetLiftCoeffFunc (0); 
	SetTrimScale (0.05);
	SetCameraOffset(CAMERA_OFS);

	
	/* CVE-lite function ! */
	/* IMPORTANT */
	/* Draw() should be called to render all the loaded payloads, plus anything you've
	/* added with the COMMON_MESH class (not used here) */

	/* call draw to render the payloads, plus anything you've added from COMMON_MESH */
	PlaceComponents();
	Draw();
	SetMeshVisibilityMode(3,MESHVIS_ALWAYS);
	SetMeshVisibilityMode(0,MESHVIS_EXTERNAL|MESHVIS_VC);
	SetMeshVisibilityMode(6,MESHVIS_EXTERNAL|MESHVIS_VC);


	/* we're not using COMMON_MESH today, so draw our own */
//	AddMesh(hX20, &VECTOR_ZERO );
	

	/* CVE-lite variable ! */
	/* IMPORTANT */
	/* Use the keyword double InitFuelLevel multipled by the maximum mass to have the
	/* fuel levels automatically set from the GPDS payload string */

	/* verify that the propellent hasn't been created before making it */
	if (ph_first==NULL)  
		ph_first  = CreatePropellantResource(FIRST_PROP_MASS,FIRST_PROP_MASS*InitFuelLevel,1.0); 
	if (!ph_deorbit)  
		ph_deorbit  = CreatePropellantResource(DEORBIT_PROP_MASS,DEORBIT_PROP_MASS,1.0); 
//	if (!ph_x20)  
//		ph_x20  = CreatePropellantResource(RCS_PROP_MASS,RCS_PROP_MASS,1.0); 

	SetDefaultPropellantResource (ph_first); 

	th_first[0] = CreateThruster ( _V(0,0.0,-9.2), _V(0,0,1), THRUST_CHARIOT_VAC/2, ph_first, ISP_CHARIOT_VAC, ISP_CHARIOT_SL);
//	AddExhaustStream (th_first[0], _V(0.72,0,-7), &contrail);
//	AddExhaustStream (th_first[0], _V(0.72,0,-7), &exhaust_main);

//	th_first[1] = CreateThruster ( _V(0,-0.72,-9.2), _V(0,0,1), THRUST_CHARIOT_VAC/2, ph_first, ISP_CHARIOT_VAC, ISP_CHARIOT_SL);
//	AddExhaustStream (th_first[1], _V(-0.72,0,-7), &contrail);
//	AddExhaustStream (th_first[1], _V(-0.72,0,-7), &exhaust_main);

	thg_first = CreateThrusterGroup (th_first, 1, THGROUP_MAIN);
	
	AddExhaust (th_first[0], 11.0, 1.1, tex);

	SetAttControlsTranstage();
	EnableTransponder(true);

}

void X20::SetDeorbitStage( )
{


	/* CVE-lite function ! */
	/* ClearMeshesEx() will get rid of all the normal meshes, plus anything you've defined
	/* with the COMMON_MESH class (not used here) */

	/* erase all meshes and thrusters so we can start afresh */
	ClearMeshesEx();
	ClearThrusterDefinitions();

/*	PARTICLESTREAMSPEC contrail = 
	{
		0, 5.0, 5, 150, 0.3, 4.0, 16, 3.0, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};
*/
	PARTICLESTREAMSPEC exhaust_main = {
		0, 0.5, 3, 500, 0, 1, 50, 0.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_FLAT, 1, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 0.3, 0.3
	};

	exhaust_main.tex = oapiRegisterParticleTexture ("Contrail2");
	SetSize (14);

	/* CVE-lite function ! */
	/* IMPORTANT */
	/* double CalculateMass() will return the all-up mass of the combined payloads and
	/* upper stages. Simply add this to your normal dry mass. */
	
	/* always add payload mass [CalculateMass()] to the vessel's empty mass */
	SetEmptyMass ( X20_DRYMASS + DEORBIT_DRYMASS + atchmass + CalculateMass() );


	SetTouchdownPoints (_V(0,-1.0,-3), _V(-.7,.7,-3), _V(.7,.7,-3));
	SetPMI (_V(16,17.5,3.0));
	SetCrossSections (_V(32.4,56.4,12.5));
	SetCW (0.1, 0.3, 1.4, 1.4);
	SetRotDrag (_V(0.7,0.7,1.2));
	SetPitchMomentScale (1e-6);
	SetBankMomentScale (1e-6);
//	SetLiftCoeffFunc (0); 
	SetTrimScale (0.05);
	SetCameraOffset(CAMERA_OFS);

	
	/* CVE-lite function ! */
	/* IMPORTANT */
	/* Draw() should be called to render all the loaded payloads, plus anything you've
	/* added with the COMMON_MESH class (not used here) */

	/* call draw to render the payloads, plus anything you've added from COMMON_MESH */
	PlaceComponents();
	Draw();
	SetMeshVisibilityMode(3,MESHVIS_ALWAYS);
	SetMeshVisibilityMode(0,MESHVIS_EXTERNAL|MESHVIS_VC);
	SetMeshVisibilityMode(6,MESHVIS_EXTERNAL|MESHVIS_VC);


	/* we're not using COMMON_MESH today, so draw our own */
//	AddMesh(hX20, &VECTOR_ZERO );
	

	/* CVE-lite variable ! */
	/* IMPORTANT */
	/* Use the keyword double InitFuelLevel multipled by the maximum mass to have the
	/* fuel levels automatically set from the GPDS payload string */

	/* verify that the propellent hasn't been created before making it */

	if (ph_first != NULL)
		DelPropellantResource (ph_first);
	if (ph_deorbit==NULL)  
		ph_deorbit  = CreatePropellantResource(DEORBIT_PROP_MASS,DEORBIT_PROP_MASS,1.0); 
	if (ph_x20==NULL)  
		ph_x20  = CreatePropellantResource(RCS_PROP_MASS,RCS_PROP_MASS,1.0); 

	SetDefaultPropellantResource (ph_deorbit); 

	th_first[0] = CreateThruster ( _V(-0.271,-0.136,-2.8), _V(0,0.1472,1), THRUST_DEORBIT_VAC, ph_deorbit, ISP_DEORBIT_VAC, ISP_DEORBIT_SL);
	th_first[1] = CreateThruster ( _V( 0.271,-0.136,-2.8), _V(0,0.1472,1), THRUST_DEORBIT_VAC, ph_deorbit, ISP_DEORBIT_VAC, ISP_DEORBIT_SL);
	th_first[2] = CreateThruster ( _V(-0.271,-0.688,-2.8), _V(0,0.1472,1), THRUST_DEORBIT_VAC, ph_deorbit, ISP_DEORBIT_VAC, ISP_DEORBIT_SL);
	th_first[3] = CreateThruster ( _V( 0.271,-0.688,-2.8), _V(0,0.1472,1), THRUST_DEORBIT_VAC, ph_deorbit, ISP_DEORBIT_VAC, ISP_DEORBIT_SL);

//	stream_a_1 = AddExhaustStream (th_first[0], _V(-0.271,-0.136,-2.8), &contrail);
	stream_b_1 = AddExhaustStream (th_first[0], _V(-0.271,-0.136,-2.8), &exhaust_main);
//	stream_a_2 = AddExhaustStream (th_first[1], _V( 0.271,-0.136,-2.8), &contrail);
	stream_b_2 = AddExhaustStream (th_first[1], _V( 0.271,-0.136,-2.8), &exhaust_main);
//	stream_a_3 = AddExhaustStream (th_first[2], _V(-0.271,-0.688,-2.8), &contrail);
	stream_b_3 = AddExhaustStream (th_first[2], _V(-0.271,-0.688,-2.8), &exhaust_main);
//	stream_a_4 = AddExhaustStream (th_first[3], _V( 0.271,-0.688,-2.8), &contrail);
	stream_b_4 = AddExhaustStream (th_first[3], _V( 0.271,-0.688,-2.8), &exhaust_main);

	thg_first = CreateThrusterGroup (th_first, 4, THGROUP_MAIN);


	Exh1 = AddExhaust (th_first[0], 2.0, 0.3, tex);
	Exh2 = AddExhaust (th_first[1], 2.0, 0.3, tex);
	Exh3 = AddExhaust (th_first[2], 2.0, 0.3, tex);
	Exh4 = AddExhaust (th_first[3], 2.0, 0.3, tex);

	SetAttControlsX20();
	EnableTransponder(true);
}

void X20::SetX20Stage( )
{


	/* CVE-lite function ! */
	/* ClearMeshesEx() will get rid of all the normal meshes, plus anything you've defined
	/* with the COMMON_MESH class (not used here) */

	/* erase all meshes and thrusters so we can start afresh */
	ClearMeshesEx();
	ClearThrusterDefinitions();

	PARTICLESTREAMSPEC contrail = 
	{
		0, 25.0, 7, 350, 0.3, 4.0, 7, 3.0, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};

	PARTICLESTREAMSPEC exhaust_main = {
		0, 3, 25, 350, 0.1, 0.2, 28, 1.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-5, 0.1
	};
	exhaust_main.tex = oapiRegisterParticleTexture ("Contrail2");
	SetSize (12.5);

	/* CVE-lite function ! */
	/* IMPORTANT */
	/* double CalculateMass() will return the all-up mass of the combined payloads and
	/* upper stages. Simply add this to your normal dry mass. */
	
	/* always add payload mass [CalculateMass()] to the vessel's empty mass */
	SetEmptyMass ( X20_DRYMASS + atchmass + CalculateMass() );


	SetTouchdownPoints (_V(0,-1.0,-3), _V(-.7,.7,-3), _V(.7,.7,-3));
	SetPMI (_V(16.7,18.9,3.6));
	SetCrossSections (_V(24.2,55.4,9.8));
	SetCW (0.1, 0.3, 1.4, 1.4);
	SetRotDrag (_V(0.7,0.7,1.2));
	SetPitchMomentScale (1e-6);
	SetBankMomentScale (1e-6);
//	SetLiftCoeffFunc (0); 
	SetTrimScale (0.05);
	SetCameraOffset(CAMERA_OFS);

	
	/* CVE-lite function ! */
	/* IMPORTANT */
	/* Draw() should be called to render all the loaded payloads, plus anything you've
	/* added with the COMMON_MESH class (not used here) */

	/* call draw to render the payloads, plus anything you've added from COMMON_MESH */
	PlaceComponents();
	Draw();
	SetMeshVisibilityMode(3,MESHVIS_ALWAYS);
	SetMeshVisibilityMode(0,MESHVIS_EXTERNAL|MESHVIS_VC);
	SetMeshVisibilityMode(6,MESHVIS_EXTERNAL|MESHVIS_VC);

	/* we're not using COMMON_MESH today, so draw our own */
//	AddMesh(hX20, &VECTOR_ZERO );
	

	/* CVE-lite variable ! */
	/* IMPORTANT */
	/* Use the keyword double InitFuelLevel multipled by the maximum mass to have the
	/* fuel levels automatically set from the GPDS payload string */

	/* verify that the propellent hasn't been created before making it */

	if (ph_first != NULL)
		DelPropellantResource (ph_first);
	if (ph_deorbit != NULL)				
		DelPropellantResource (ph_deorbit); 
	if (ph_x20==NULL)  
		ph_x20  = CreatePropellantResource(RCS_PROP_MASS,RCS_PROP_MASS,1.0); 

	SetDefaultPropellantResource (ph_x20); 
/*
	th_first[0] = CreateThruster ( _V(0,0.72,-2.8), _V(0,0,1), THRUST_FIRST_VAC, ph_first, ISP_FIRST_VAC, ISP_FIRST_SL);
//	AddExhaustStream (th_first[0], _V(0.72,0,-7), &contrail);
//	AddExhaustStream (th_first[0], _V(0.72,0,-7), &exhaust_main);

	th_first[1] = CreateThruster ( _V(0,-0.72,-2.8), _V(0,0,1), THRUST_FIRST_VAC, ph_first, ISP_FIRST_VAC, ISP_FIRST_SL);
//	AddExhaustStream (th_first[1], _V(-0.72,0,-7), &contrail);
//	AddExhaustStream (th_first[1], _V(-0.72,0,-7), &exhaust_main);

	thg_first = CreateThrusterGroup (th_first, 2, THGROUP_MAIN);
	
	AddExhaust (th_first[0], 8.0, 0.9, tex);
	AddExhaust (th_first[1], 8.0, 0.9, tex);
*/
	SetAttControlsX20();
	EnableTransponder(true);

}



/** Handle buffered key events sent to the vessel by orbiter. 
 *  @param dwKey The scan code of the key
 *  @param fPressed true if pressed, false if depressed
 *  @param keystate the current state of the keyboard
 *
 *  modified switch routine by dave rowbotham.
 */

void X20::JettisonPart()
 {
				VESSELSTATUS vs;
				VECTOR3 ofs = _V(0,0,0);
				VECTOR3 vel = _V(0,0,-0.4);
//				ofs = Centaur_OFS;
				GetStatus (vs);
				vs.eng_main = vs.eng_hovr = 0.0;
				SetEngineLevel(ENGINE_MAIN, 0);
				VECTOR3 rofs, rvel = {vs.rvel.x, vs.rvel.y, vs.rvel.z};
				Local2Rel (ofs, vs.rpos);
				GlobalRot (vel, rofs);
				vs.rvel.x = rvel.x+rofs.x;
				vs.rvel.y = rvel.y+rofs.y;
				vs.rvel.z = rvel.z+rofs.z;
				vs.vrot.x = 0.0;
				vs.vrot.y = 0.0;
				vs.vrot.z = 0.0;
				char name[256];
				if (stage == 0) {
					if (stagetype == 1) {
						strcpy (name, GetName()); strcat (name, "-Transtage");
						VESSEL::Create (name, "X20/X20_Transtage", vs);	
					} else if (stagetype == 2) {
						strcpy (name, GetName()); strcat (name, "-Centaur");
						VESSEL::Create (name, "X20/X20_Centaur", vs);	
					} else if (stagetype == 3) {
						strcpy (name, GetName()); strcat (name, "-Chariot");
						VESSEL::Create (name, "X20/X20_Chariot", vs);	
					}
					DelPropellantResource (ph_first); 
					PlayVesselWave3(MyID,PLAYBANG);
				}
				else if (GetThrusterGroupLevel (thg_first) == 0.0) {
					strcpy (name, GetName()); strcat (name, "-Deorbit Stage");
					VESSEL::Create (name, "X20/X20_Deorbit", vs);
					DelPropellantResource (ph_deorbit); 
					PlayVesselWave3(MyID,PLAYBANG);
				}
				else stage--;


				stage++;

 }

void X20::AttachSweep()
{

	
			/* Now scan for nearby attach points; these need to stay in the bay!  */
				
			Local2Global (_V(0,0,4.8), grms);  // global position of RMS tip
			
			// Search the complete vessel list for a grappling candidate.
			// Not very scalable ...

			for (i = 0; i < oapiGetVesselCount(); i++) {
				hV = oapiGetVesselByIndex (i);
				if (hV == GetHandle()) continue; // we don't want to grapple ourselves ...
				oapiGetGlobalPos (hV, &gpos);
//				sprintf(oapiDebugString(),"Distance is, %d, Too far to strap in!",dist(gpos, grms));
				if (dist (gpos, grms) < oapiGetSize (hV)) { // in range
					VESSEL *v = oapiGetVesselInterface (hV);
					DWORD nAttach = v->AttachmentCount (true);
					for (DWORD j = 0; j < nAttach; j++) { // now scan all attachment points of the candidate
						ATTACHMENTHANDLE hAtt = v->GetAttachmentHandle (true, j);
						const char *id = v->GetAttachmentId (hAtt);
						if (strncmp (id, "BOLT", 4)) continue; // attachment point not compatible
						v->GetAttachmentParams (hAtt, pos, dir, rot);
						v->Local2Global (pos, gpos);
						if (dist (gpos, grms) < 0.25) { // found one!
							// check whether satellite is currently clamped into payload bay
							if ((hV == GetAttachmentStatus (Baypos)) && (!Firstsweep))
							{
								AttachChild (hV, Baypos, hAtt);
								atchmass = v->GetMass();
								BayVessel = hV;
								BayV = oapiGetVesselInterface (BayVessel);
							}
							else
							{	
								AttachChild (hV, Baypos, hAtt);
								atchmass = v->GetMass();
								hDock = v->GetDockHandle(0);
								BayVessel = hV;
								BayV = oapiGetVesselInterface (BayVessel);
								if (hDock != NULL) {
									v->GetDockParams(hDock,pos, dir, rot);
									pos.z += 4.8;
//									CreateDock(pos, dir, rot);
									SetDockParams(Dock0, pos, dir, rot);
									v->DelDock(hDock);
								}
								else {
									Undock(ALLDOCKS);
									DelDock(Dock0);
								}
							}
//							return 1;
						}
					}
				}
			}
}

int X20::ConsumeBufferedKey(DWORD dwKey, bool fPressed, const char *keystate)
{
	if (!fPressed) return 0; // only process keydown events
	if (KEYMOD_SHIFT (keystate)) 
	{
	}
	else if (KEYMOD_CONTROL (keystate)) 
	{
		// insert ctrl key combinations here
		if (KEYDOWN (keystate, OAPI_KEY_E))
		{ 
			if (oapiAcceptDelayedKey (OAPI_KEY_E, 1.0))
				// Report the inclination setpoint
				// Eject Eject Eject!!!!

			if (!bEjected) {

				PlayVesselWave3(MyID,PLAYBANG);
				bEjected = true;
				VESSELSTATUS vs;
				GetStatus (vs);
				vs.eng_main = vs.eng_hovr = 0.0;
				vs.fuel=1.0;
				vs.status=0;
				VECTOR3 ofs = TOPHATCH_OFS;
				VECTOR3 vel = _V(0,20.0,-5.0);
				VECTOR3 rofs, rvel = {vs.rvel.x, vs.rvel.y, vs.rvel.z};
				Local2Rel (ofs, vs.rpos);
				GlobalRot (vel, rofs);
				vs.rvel.x = rvel.x+rofs.x;
				vs.rvel.y = rvel.y+rofs.y;
				vs.rvel.z = rvel.z+rofs.z;
				vs.vrot.x = 0.7;
				vs.vrot.y = 0.0;
				vs.vrot.z = 0.0;
				char name[256];
				strcpy (name, GetName()); strcat (name, "-Hatch");
				VESSEL::Create (name, "X20/X20_TopHatch", vs);

				GetStatus (vs);
				ofs = ASTRO_OFS;
				vel = _V(0,12.0,0.0);
	//			rofs = {vs.rvel.x, vs.rvel.y, vs.rvel.z};
	//			rvel = {vs.rvel.x, vs.rvel.y, vs.rvel.z};
				Local2Rel (ofs, vs.rpos);
				GlobalRot (vel, rofs);
				vs.rvel.x = rvel.x+rofs.x;
				vs.rvel.y = rvel.y+rofs.y;
				vs.rvel.z = rvel.z+rofs.z;
				vs.vrot.x = 0.0;
				vs.vrot.y = 0.0;
				vs.vrot.z = 0.0;
	//			char name[256];
				strcpy (name, GetName()); strcat (name, "-Pilot");
				vs.fuel = 1.0;
				VESSEL::Create (name, "X20/MOOSE", vs);
				OBJHANDLE hvessel;
				hvessel=oapiGetVesselByName(name);
				if (hvessel != 0) {
					VESSEL *hv;
					hv = oapiGetVesselInterface(hvessel);
					hv->SetThrusterGroupLevel (THGROUP_HOVER,1.0);
					oapiSetFocusObject(hvessel);
				}
				SetMainStage();
			}
																
			return 1;
		}
		return 0;
	}
	else { // unmodified keys
		if (KEYDOWN (keystate, OAPI_KEY_J)) { // "Jettison Stage"
			if (oapiAcceptDelayedKey (OAPI_KEY_J, 1.0)) {
				if ((bThrustTerminated) && (GetFuelMass() > 0.0)) return 0;

					if ((stage < 2) && (pldoors_proc == 0.0)) {
						JettisonPart();
					}
					else {
						OverridePayloadVelocity(_V(0,+0.25,0));
						DeployPayload();
					}

					/* CVE-lite function ! */
					/* IMPORTANT */
					/* call DeployPayload() to jettison the GPDS payloads in sequence */

					AttachSweep();

					/* IMPORTANT */
					/* Always re-set the stage after any kind of change */
					SetMainStage();
				
			}
			return 1;
		}
		if (KEYDOWN (keystate, OAPI_KEY_G)) {
			if (oapiAcceptDelayedKey (OAPI_KEY_G, 1.0)) {
				// Toggle the Gear
				RevertLandingGear ();
				StopVesselWave3(MyID,PLAYGEAR);
				PlayVesselWave3(MyID,PLAYGEAR);
			}
			return 1;
		}
		if (KEYDOWN (keystate, OAPI_KEY_K)) {
			if (oapiAcceptDelayedKey (OAPI_KEY_K, 1.0)) {
				// Toggle the bay doors
				RevertPayloadBayDoors ();
				if (BayVessel != NULL)
					SendCentralMessage(BayV, "MSG", "REVERTPROTRUSIONS");
				if (pldoors_status == DOOR_CLOSING)
					Undock(ALLDOCKS);
			}
			return 1;
		}
		if (KEYDOWN (keystate, OAPI_KEY_B)) {
			if (oapiAcceptDelayedKey (OAPI_KEY_B, 1.0)) {
				// Toggle the brakes
				if (BrakesOn)
					BrakesOn = false;
				else BrakesOn = true;
				RevertBrakes();
			}
			return 1;
		}
		if (KEYDOWN (keystate, OAPI_KEY_P)) {
			if (oapiAcceptDelayedKey (OAPI_KEY_P, 1.0)) {
				// Toggle the top hatch
				if (!bEjected)
					RevertHatch ();
			}
			return 1;
		}
		if (KEYDOWN (keystate, OAPI_KEY_O)) {
			if (oapiAcceptDelayedKey (OAPI_KEY_O, 1.0)) {
				// Disable the Autopilot
				AutoPilot = false;
				if (stage == 0) {
					ActivateNavmode(NAVMODE_KILLROT);
					SetAttitudeRotLevel( 0 , 0.0);
					SetAttitudeRotLevel( 1 , 0.0);
					SetAttitudeRotLevel( 2 , 0.0);
					SetThrusterGroupLevel (thg_first,0.0);			
				}
			}
			return 1;
		}
		if (KEYDOWN (keystate, OAPI_KEY_E)) {
			if (oapiAcceptDelayedKey (OAPI_KEY_E, 1.0)) {
				// EVA
				if ((hatch_status == DOOR_OPEN) && (!bEjected))
				{
					char name[256];
					OBJHANDLE hvessel;
					strcpy (name, GetName()); strcat (name, "-PLT");
					hvessel=oapiGetVesselByName(name);
					if (hvessel == 0)
					{
						bEVA = true;
						SeparateEVA();
						SetMainStage();
						hvessel=oapiGetVesselByName(name);
						if (hvessel != 0)
							oapiSetFocusObject(hvessel);
					}
					else
					{
						oapiDeleteVessel(hvessel);
						bEVA = false;
						SetMainStage();
					};
				}

			}
			return 1;
		}

	}
	return 0;
}



int X20::ConsumeKey( const char *keystate )
{

	if (KEYMOD_SHIFT (keystate)) 
	{
						
		return 0; // shift combinations are reserved
	}
	else if (KEYMOD_CONTROL (keystate)) 
	{
		return 0;
	}
	else 
	{ // unmodified keys
	
		return 0;

	}
	return 0;
}




void X20::SetAttControlsTranstage()
{
	THRUSTER_HANDLE th_rot[2];
	texRCS  = oapiRegisterExhaustTexture ("exhaust_atrcs");

	double zofs = 0.0;
	if (stagetype == 2) zofs = -2.0;

	th_rot[0] = CreateThruster (_V(0,0, 5), _V(0, 1,0), MAX_ATT_FIRST, ph_first, ISP_FIRST_SL);
	th_rot[1] = CreateThruster (_V(0,0,-5), _V(0,-1,0), MAX_ATT_FIRST, ph_first, ISP_FIRST_SL);
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 1.7, 0.25,-5.5+zofs), _V(0,1.0,0.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V(-1.7, 0.25,-5.5+zofs), _V(0,1.0,0.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 1.18,-1.27,-0.35), _V(0,-1.0,0.0),texRCS);  // X-20
	AddExhaust (th_rot[0], 1.5, 0.35, _V(-1.18,-1.27,-0.35), _V(0,-1.0,0.0),texRCS);  // X-20
	thg_pitchup=CreateThrusterGroup (th_rot, 2, THGROUP_ATT_PITCHUP);

	th_rot[0] = CreateThruster (_V(0,-0.2,-5.25), _V(0,1,0),MAX_ATT_FIRST, ph_first, ISP_FIRST_SL);
	th_rot[1] = CreateThruster (_V(0,-0.75,5.25), _V(0,1,0),MAX_ATT_FIRST, ph_first, ISP_FIRST_SL);
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 1.7,-0.187,-5.35+zofs), _V(0,-1.0,0.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V(-1.7,-0.187,-5.35+zofs), _V(0,-1.0,0.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 1.18,-1.27,-0.35), _V(0,-1.0,0.0),texRCS);  // X-20
	AddExhaust (th_rot[0], 1.5, 0.35, _V(-1.18,-1.27,-0.35), _V(0,-1.0,0.0),texRCS);  // X-20
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_UP);

	th_rot[0] = CreateThruster (_V(0,-0.2,-5.25), _V(0,-1,0), MAX_ATT_FIRST, ph_first, ISP_FIRST_SL);
	th_rot[1] = CreateThruster (_V(0,-0.75,5.25), _V(0,-1,0), MAX_ATT_FIRST, ph_first, ISP_FIRST_SL);
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 1.7, 0.25,-5.5+zofs), _V(0,1.0,0.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V(-1.7, 0.25,-5.5+zofs), _V(0,1.0,0.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 0.8, 0.25, _V( 2.15,-1.0, 3.94), _V(0, 1.0,0.0),texRCS);  // X-20
	AddExhaust (th_rot[0], 0.8, 0.25, _V(-2.15,-1.0, 3.94), _V(0, 1.0,0.0),texRCS);  // X-20
	AddExhaust (th_rot[0], 0.8, 0.25, _V( 2.30,-1.0, 3.80), _V(0, 1.0,0.0),texRCS);  // X-20
	AddExhaust (th_rot[0], 0.8, 0.25, _V(-2.30,-1.0, 3.80), _V(0, 1.0,0.0),texRCS);  // X-20
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_DOWN);

	th_rot[0] = CreateThruster (_V(0,0, 5), _V(0,-1,0), MAX_ATT_FIRST, ph_first, ISP_FIRST_SL);
	th_rot[1] = CreateThruster (_V(0,0,-5), _V(0, 1,0), MAX_ATT_FIRST, ph_first, ISP_FIRST_SL);
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 1.7,-0.187,-5.35+zofs), _V(0,-1.0,0.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V(-1.7,-0.187,-5.35+zofs), _V(0,-1.0,0.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 0.8, 0.25, _V( 2.15,-1.0, 3.94), _V(0, 1.0,0.0),texRCS);  // X-20
	AddExhaust (th_rot[0], 0.8, 0.25, _V(-2.15,-1.0, 3.94), _V(0, 1.0,0.0),texRCS);  // X-20
	AddExhaust (th_rot[0], 0.8, 0.25, _V( 2.30,-1.0, 3.80), _V(0, 1.0,0.0),texRCS);  // X-20
	AddExhaust (th_rot[0], 0.8, 0.25, _V(-2.30,-1.0, 3.80), _V(0, 1.0,0.0),texRCS);  // X-20
	thg_pitchdown=CreateThrusterGroup (th_rot, 2, THGROUP_ATT_PITCHDOWN);

	th_rot[0] = CreateThruster (_V(0,0, 4), _V(-1,0,0), MAX_ATT_FIRST, ph_first, ISP_FIRST_SL);
	th_rot[1] = CreateThruster (_V(0,0,-4), _V( 1,0,0), MAX_ATT_FIRST, ph_first, ISP_FIRST_SL);
	AddExhaust (th_rot[0], 1.5, 0.35, _V(-0.21, 1.7,-5.35+zofs), _V(-1.0,0,0.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V(-0.21,-1.7,-5.35+zofs), _V(-1.0,0,0.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 1.10,-0.50, 2.3), _V( 1.0,0,0.0),texRCS);  // X-20
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_YAWLEFT);

	th_rot[0] = CreateThruster (_V(0,0, 4), _V( 1,0,0), MAX_ATT_FIRST, ph_first, ISP_FIRST_SL);
	th_rot[1] = CreateThruster (_V(0,0,-4), _V(-1,0,0), MAX_ATT_FIRST, ph_first, ISP_FIRST_SL);
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 0.25, 1.7,-5.55+zofs), _V( 1.0,0,0.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 0.25,-1.7,-5.55+zofs), _V( 1.0,0,0.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V(-1.10,-0.50, 2.3), _V(-1.0,0,0.0),texRCS);  // X-20
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_YAWRIGHT);

	th_rot[0] = CreateThruster (_V(2,0.0,0.0), _V(-1,0,0), MAX_ATT_FIRST, ph_first, ISP_FIRST_SL);
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 0.25, 1.7,-5.55+zofs), _V( 1.0,0,0.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 0.25,-1.7,-5.55+zofs), _V( 1.0,0,0.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 1.10,-0.50, 2.3), _V( 1.0,0,0.0),texRCS);  // X-20
	CreateThrusterGroup (th_rot, 1, THGROUP_ATT_LEFT);

	th_rot[0] = CreateThruster (_V(2,0.0,0.0), _V(1,0,0),MAX_ATT_FIRST, ph_first, ISP_FIRST_SL);
	AddExhaust (th_rot[0], 1.5, 0.35, _V(-0.21, 1.7,-5.35+zofs), _V(-1.0,0,0.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V(-0.21,-1.7,-5.35+zofs), _V(-1.0,0,0.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V(-1.10,-0.50, 2.3), _V(-1.0,0,0.0),texRCS);  // X-20
	CreateThrusterGroup (th_rot, 1, THGROUP_ATT_RIGHT);

	th_rot[0] = CreateThruster (_V(13,0,0), _V(0,-1,0), MAX_ATT_FIRST, ph_first, ISP_FIRST_SL);
	th_rot[1] = CreateThruster (_V(-13,0,0), _V(0, 1,0), MAX_ATT_FIRST, ph_first, ISP_FIRST_SL);
	AddExhaust (th_rot[0], 1.5, 0.35, _V(-0.21, 1.7,-5.35+zofs), _V(-1.0,0,0.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 0.25,-1.7,-5.55+zofs), _V( 1.0,0,0.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 1.7, 0.25,-5.5+zofs), _V(0,1.0,0.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V(-1.7,-0.187,-5.35+zofs), _V(0,-1.0,0.0),texRCS);  // Transtage
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_BANKRIGHT);

	th_rot[0] = CreateThruster (_V(13,0, 0), _V(0, 1,0), MAX_ATT_FIRST, ph_first, ISP_FIRST_SL);
	th_rot[1] = CreateThruster (_V(-13,0,0), _V(0, -1,0), MAX_ATT_FIRST, ph_first, ISP_FIRST_SL);
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 0.25, 1.7,-5.55+zofs), _V( 1.0,0,0.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V(-0.21,-1.7,-5.35+zofs), _V(-1.0,0,0.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 1.7,-0.187,-5.35+zofs), _V(0,-1.0,0.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V(-1.7, 0.25,-5.5+zofs), _V(0,1.0,0.0),texRCS);  // Transtage
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_BANKLEFT);

	th_rot[0] = CreateThruster (_V(0,0, -4), _V(0,0,1), MAX_ATT_FIRST, ph_first, ISP_FIRST_SL);
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 1.7, 0.0,-5.75+zofs), _V(0,0,-1.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V(-1.7, 0.0,-5.75+zofs), _V(0,0,-1.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 0.0, 1.7,-5.75+zofs), _V(0,0,-1.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 0.0,-1.7,-5.75+zofs), _V(0,0,-1.0),texRCS);  // Transtage
	CreateThrusterGroup (th_rot, 1, THGROUP_ATT_FORWARD);
	th_rot[0] = CreateThruster (_V(0,0, 4), _V(0,0,-1), MAX_ATT_FIRST, ph_first, ISP_FIRST_SL);
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 1.7, 0.0,-5.15+zofs), _V(0,0,1.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V(-1.7, 0.0,-5.15+zofs), _V(0,0,1.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 0.0, 1.7,-5.15+zofs), _V(0,0,1.0),texRCS);  // Transtage
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 0.0,-1.7,-5.15+zofs), _V(0,0,1.0),texRCS);  // Transtage
	CreateThrusterGroup (th_rot, 1, THGROUP_ATT_BACK);
}


void X20::SetAttControlsX20()
{
	THRUSTER_HANDLE th_rot[2];
	th_rot[0] = CreateThruster (_V(0,0, 5), _V(0, 1,0), MAX_ATT_FIRST, ph_x20, ISP_FIRST_SL);
	th_rot[1] = CreateThruster (_V(0,0,-5), _V(0,-1,0), MAX_ATT_FIRST, ph_x20, ISP_FIRST_SL);
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 0.2, 1.3, 0.4), _V(0, 1.0,0.0),texRCS);  // X-20
	AddExhaust (th_rot[0], 1.5, 0.35, _V(-0.2, 1.3, 0.4), _V(0, 1.0,0.0),texRCS);  // X-20
	thg_pitchup=CreateThrusterGroup (th_rot, 2, THGROUP_ATT_PITCHUP);

//	th_rot[0] = CreateThruster (_V(0,-0.2,-5.25), _V(0,1,0),MAX_ATT_FIRST, ph_x20, ISP_FIRST_SL);
//	th_rot[1] = CreateThruster (_V(0,-0.75,5.25), _V(0,1,0),MAX_ATT_FIRST, ph_x20, ISP_FIRST_SL);
//	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_UP);
	th_rot[0] = CreateThruster (_V(0,-0.2,-5.25), _V(0,-1,0), MAX_ATT_FIRST, ph_x20, ISP_FIRST_SL);
	th_rot[1] = CreateThruster (_V(0,-0.75,5.25), _V(0,-1,0), MAX_ATT_FIRST, ph_x20, ISP_FIRST_SL);
	AddExhaust (th_rot[0], 0.8, 0.25, _V( 2.15,-1.0, 3.94), _V(0, 1.0,0.0),texRCS);  // X-20
	AddExhaust (th_rot[0], 0.8, 0.25, _V(-2.15,-1.0, 3.94), _V(0, 1.0,0.0),texRCS);  // X-20
	AddExhaust (th_rot[0], 0.8, 0.25, _V( 2.30,-1.0, 3.80), _V(0, 1.0,0.0),texRCS);  // X-20
	AddExhaust (th_rot[0], 0.8, 0.25, _V(-2.30,-1.0, 3.80), _V(0, 1.0,0.0),texRCS);  // X-20
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_DOWN);

	th_rot[0] = CreateThruster (_V(0,0, 5), _V(0,-1,0), MAX_ATT_FIRST, ph_x20, ISP_FIRST_SL);
	th_rot[1] = CreateThruster (_V(0,0,-5), _V(0, 1,0), MAX_ATT_FIRST, ph_x20, ISP_FIRST_SL);
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 1.18,-1.27,-0.35), _V(0,-1.0,0.0),texRCS);  // X-20
	AddExhaust (th_rot[0], 1.5, 0.35, _V(-1.18,-1.27,-0.35), _V(0,-1.0,0.0),texRCS);  // X-20
	thg_pitchdown=CreateThrusterGroup (th_rot, 2, THGROUP_ATT_PITCHDOWN);

	th_rot[0] = CreateThruster (_V(0,0, 4), _V(-1,0,0), MAX_ATT_FIRST, ph_x20, ISP_FIRST_SL);
	th_rot[1] = CreateThruster (_V(0,0,-4), _V( 1,0,0), MAX_ATT_FIRST, ph_x20, ISP_FIRST_SL);
	AddExhaust (th_rot[0], 1.5, 0.35, _V(-1.10,-0.50, 2.3), _V(-1.0,0,0.0),texRCS);  // X-20
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_YAWLEFT);

	th_rot[0] = CreateThruster (_V(0,0, 4), _V( 1,0,0), MAX_ATT_FIRST, ph_x20, ISP_FIRST_SL);
	th_rot[1] = CreateThruster (_V(0,0,-4), _V(-1,0,0), MAX_ATT_FIRST, ph_x20, ISP_FIRST_SL);
	AddExhaust (th_rot[0], 1.5, 0.35, _V( 1.10,-0.50, 2.3), _V( 1.0,0,0.0),texRCS);  // X-20
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_YAWRIGHT);

//	th_rot[0] = CreateThruster (_V(2,0.0,0.0), _V(-1,0,0), MAX_ATT_FIRST, ph_x20, ISP_FIRST_SL);
//	CreateThrusterGroup (th_rot, 1, THGROUP_ATT_LEFT);
//	th_rot[0] = CreateThruster (_V(2,0.0,0.0), _V(1,0,0),MAX_ATT_FIRST, ph_x20, ISP_FIRST_SL);
//	CreateThrusterGroup (th_rot, 1, THGROUP_ATT_RIGHT);

	th_rot[0] = CreateThruster (_V(13,0,0), _V(0,-1,0), MAX_ATT_FIRST, ph_x20, ISP_FIRST_SL);
	th_rot[1] = CreateThruster (_V(-13,0,0), _V(0, 1,0), MAX_ATT_FIRST, ph_x20, ISP_FIRST_SL);
	AddExhaust (th_rot[0], 0.8, 0.25, _V( 2.15,-1.0, 3.94), _V(0, 1.0,0.0),texRCS);  // X-20
	AddExhaust (th_rot[0], 0.8, 0.25, _V( 2.30,-1.0, 3.80), _V(0, 1.0,0.0),texRCS);  // X-20
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_BANKRIGHT);

	th_rot[0] = CreateThruster (_V(13,0, 0), _V(0, 1,0), MAX_ATT_FIRST, ph_x20, ISP_FIRST_SL);
	th_rot[1] = CreateThruster (_V(-13,0,0), _V(0, -1,0), MAX_ATT_FIRST, ph_x20, ISP_FIRST_SL);
	AddExhaust (th_rot[0], 0.8, 0.25, _V(-2.15,-1.0, 3.94), _V(0, 1.0,0.0),texRCS);  // X-20
	AddExhaust (th_rot[0], 0.8, 0.25, _V(-2.30,-1.0, 3.80), _V(0, 1.0,0.0),texRCS);  // X-20
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_BANKLEFT);

	th_rot[0] = CreateThruster (_V(0,0, -4), _V(0,0,1), MAX_ATT_FIRST, ph_x20, ISP_FIRST_SL);
	if (stage == 1) {
		AddExhaust (th_rot[0], 0.8, 0.25, _V( 1.4, 0.0,-2.88), _V(0, 0,-1.0),texRCS);  // X-20
		AddExhaust (th_rot[0], 0.8, 0.25, _V(-1.4, 0.0,-2.88), _V(0, 0,-1.0),texRCS);  // X-20
	}
	else
		AddExhaust (th_rot[0], 0.8, 0.25, _V( 0.0,0.35, 0.23), _V(0, 0,-1.0),texRCS);  // X-20
	CreateThrusterGroup (th_rot, 1, THGROUP_ATT_FORWARD);
//	th_rot[0] = CreateThruster (_V(0,0, 4), _V(0,0,-1), MAX_ATT_FIRST, ph_x20, ISP_FIRST_SL);
//	CreateThrusterGroup (th_rot, 1, THGROUP_ATT_BACK);
}


bool X20::clbkLoadVC (int id)
{
//	int i;

//	ReleaseSurfaces();
//	InitVC (id);
	SetCameraDefaultDirection (_V(0,0,1)); // forward
//	oapiVCRegisterHUD (&huds);
//	oapiVCRegisterMFD (MFD_LEFT, &mfds_left);
	oapiVCRegisterMFD (MFD_RIGHT, &mfds_right);


	return true;
}





/* basic vessel class setup */

void X20::clbkSetClassCaps (FILEHANDLE cfg)
{
	SetClassCaps(cfg);
}

void X20::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
	LoadState(  scn, vs );
}

void X20::clbkSaveState (FILEHANDLE scn)
{
	SaveState(  scn );
}

void X20::clbkPostStep (double simt, double simdt, double mjd)
{
	TimeStep(simt);
}

void X20::clbkVisualCreated (VISHANDLE _vis, int refcount)
{
	SetMainStage( );
}

void X20::clbkVisualDestroyed (VISHANDLE _vis, int refcount)
{
	SetMainStage( );
}

int X20::clbkConsumeBufferedKey (DWORD dwKey, bool fPressed, char *keystate)
{
	int ret = ConsumeBufferedKey( dwKey, fPressed, keystate );
	return ret;
}



bool X20::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	// show gear deployment status
	if (gear_status == DOOR_OPEN || (gear_status >= DOOR_CLOSING && fmod (oapiGetSimTime(), 1.0) < 0.5)) {
		int d = hps->Markersize/2;
		int cx = hps->CX, cy = hps->CY;
		if (cx >= -d*3 && cx < hps->W+d*3 && cy >= d && cy < hps->H+d*5) {
			skp->Rectangle (cx-d/2, cy-d*5, cx+d/2, cy-d*4);
			skp->Rectangle (cx-d*3, cy-d*2, cx-d*2, cy-d);
			skp->Rectangle (cx+d*2, cy-d*2, cx+d*3, cy-d);
		}
	}


	// show pldoors deployment status
	if (pldoors_status == DOOR_OPEN || (pldoors_status >= DOOR_CLOSING && fmod (oapiGetSimTime(), 1.0) < 0.5)) {
		int d = hps->Markersize/2;
		int cx = hps->CX, cy = hps->CY;
		if (cx >= -d*3 && cx < hps->W+d*3 && cy >= d && cy < hps->H+d*5) {
//			skp->Arc (cx, cy-d*8.5, cx+d*4, cy-d*6.5, cx, cy-d*6, cx+d*4, cy-d*7.5);
//			skp->Arc (cx-d*4, cy-d*8.5, cx, cy-d*6.5, cx-d*4, cy-d*7.5, cx, cy-d*6);
			skp->Line(cx-d*4, cy-d*8.5, cx-d*3, cy-d*7.5);
			skp->LineTo(cx-d*2, cy-d*7.0);
			skp->LineTo(cx-d, cy-d*6.65);
			skp->LineTo(cx, cy-d*6.5);
			skp->LineTo(cx+d, cy-d*6.65);
			skp->LineTo(cx+d*2, cy-d*7.0);
			skp->LineTo(cx+d*3, cy-d*7.5);
			skp->LineTo(cx+d*4, cy-d*8.5);
			skp->Text (cx-d, cy-d*8, "BAY", 3);
		}
	}

	// show brakes deployment status
	if (BrakesOn || brakes_proc > 0) {
		int d = hps->Markersize/2;
		int cx = hps->CX, cy = hps->CY;
		if((((!BrakesOn && brakes_proc > 0) || (BrakesOn && brakes_proc < 1.0)) && 
				fmod (oapiGetSimTime(), 0.5) < 0.25)||(BrakesOn && brakes_proc == 1.0))
		{
			//Flash indicator
			if (cx >= -d*3 && cx < hps->W+d*3 && cy >= d && cy < hps->H+d*5) {
				skp->Text (cx-d*9, cy+d*1.2, "BRAKES", 6);
			}
		}
	}


	// show hatch deployment status
	if (hatch_status == DOOR_OPEN || (hatch_status >= DOOR_CLOSING && fmod (oapiGetSimTime(), 1.0) < 0.5)) {
		int d = hps->Markersize/2;
		int cx = hps->CX, cy = hps->CY;
		if (cx >= -d*3 && cx < hps->W+d*3 && cy >= d && cy < hps->H+d*5) {
			skp->Text (cx+d*5, cy+d*1.2, "DOOR", 4);
		}
	}

	//Draw Equivalent airspeed & pressure altitude
	if(GetAtmPressure() > 0.0)
	{
		int d = hps->Markersize/2;
		int cx = hps->CX, cy = hps->CY;
		char pszBuffer[40];
		int bx = 0.5 * cx;
		const double EAS_EXP = 2.0/7.0;
		const double P0 = 1.01325E6;

		double fEAS = 661.48 * sqrt(5.0 * GetAtmPressure()/ATMP * (pow(GetDynPressure() / GetAtmPressure() + 1, EAS_EXP) - 1.0));

		sprintf(pszBuffer, "%6.1f KEAS", fEAS);

		skp->Text(bx, cy - d, pszBuffer, 11);
	}
	if (GetAtmPressure() > 7000.0)				  // About 60,000'; use pressure altitude
	{
		int d = hps->Markersize/2;
		int cx = hps->CX, cy = hps->CY;
		char pszBuffer[40];
		int bx = 0.5 * cx;
		const double EAS_EXP = 2.0/7.0;
		const double P0 = 1.01325E6;

		bx = 0.5 * (cx + hps->W) - d * 8;

		double fHp = (1.0 - pow(GetAtmPressure()/ATMP, 0.190263)) * (288.15/0.00198122);

		long H_press = (long)(fHp);
		H_press = H_press - H_press % 50;

		sprintf(pszBuffer, "%6d'  ", H_press);

		skp->Text(bx, cy - d, pszBuffer, 9);
	} else {									// Do inertial altitude instead
		int d = hps->Markersize/2;		
		int cx = hps->CX, cy = hps->CY;
		char pszBuffer[40];
		int bx = 0.5 * cx;
		const double EAS_EXP = 2.0/7.0;
		const double P0 = 1.01325E6;

		bx = 0.5 * (cx + hps->W) - d * 8;

		double fIalt = GetAltitude()/0.3048;

		long I_Alt = (long)(fIalt);
		I_Alt = I_Alt - I_Alt % 50;

		sprintf(pszBuffer, "%6d' I", I_Alt);

		skp->Text(bx, cy - d, pszBuffer, 9);
	}

	//Draw Radar altitude
	if(GetAltitude() < 2500.0 * 0.3048)
	{
		int d = hps->Markersize/2;
		int cx = hps->CX, cy = hps->CY;
		char pszBuffer[40];
		int bx = 0.5 * (cx + hps->W) - d * 8;
		const double GEAR_Y = -1.35;

		sprintf(pszBuffer, "%6dR", static_cast<long>((GetAltitude() + GEAR_Y)/0.3048));

		skp->Text(bx, cy + d, pszBuffer, 7);
	}


	//Draw G-Meter

		int d = hps->Markersize/2;
		int cx = hps->CX, cy = hps->CY;

		char pszBuffer[40];
		int bx = 0.5 * cx;

		VECTOR3 ForceVec, WeightVec;
		GetForceVector(ForceVec);
		GetWeightVector(WeightVec);
		ForceVec -= WeightVec;
		double G_level = sqrt(pow(ForceVec.x,2)+pow(ForceVec.y,2)+pow(ForceVec.z,2)) / (9.81*GetMass());

		sprintf(pszBuffer, "  %.2f G", G_level);

		skp->Text(bx, cy + d, pszBuffer, strlen(pszBuffer));



	return true;
}

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new X20 (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (X20*)vessel;
}


