/**
*
* $Id: CVEL-X20.h,v 0.1 2004/11/20 18:03:45 erv Exp $
* (c) 2004 Sputnik -- erik_h_anderson@hotmail.com
* Based on CVEL Centaur-G code
* Ervins Reinverts -- ervins@gmail.com
* Based on CVEL Fregat code
* (c) 2004 David Henderson -- orbiter@aibs.org.uk
* Licence LGPL -- http://www.gnu.org/licenses/lgpl.txt
*
* You may use this source code in your addon provided that if you
* publish your addon either privately or publicly the end user
* recieves a copy of the source code including any modifications
* you may have made. 
*
*/


#include "../../include/cvel.h"
#include "../../include/orbitersdk.h"
#include "../../include/OrbiterSoundSDK3.h"
#include <cstdio>
#include <cmath>
#include <string>

#define InitFuelLevel 1

// All data from astronautix.com
const double ISP_FIRST_SL		=     316.0*G; 
const double ISP_FIRST_VAC		=     316.0*G; //***
const double ISP_DEORBIT_SL		=     231.0*G; 
const double ISP_DEORBIT_VAC	=     265.0*G; //***
const double THRUST_FIRST_VAC	=    7485.0*G; //***
const double THRUST_DEORBIT_VAC	=   18140.0*G/4.0; //***
const double MAX_ATT_FIRST		=     250.0*G;
const double FIRST_PROP_MASS	=   10297.0; //*** the total mass of fueled Transtage is 12247kg
const double DEORBIT_PROP_MASS	=	 1040.0;
const double RCS_PROP_MASS		=	  200.0;
const double X20_DRYMASS		=    5055.0; //***
const double DEORBIT_DRYMASS	=     900.0; //***
const double TRANSTAGE_DRYMASS	=    1950.0; //***
const double CENTAUR_DRYMASS	=    1700.0; //***
const double CHARIOT_DRYMASS	=    1800.0; //***
const double ISP_CENTAUR_SL		=     444.0*G; 
const double ISP_CENTAUR_VAC	=     444.0*G; //***
const double THRUST_CENTAUR_VAC	=    7485.0*G; //***
const double CENTAUR_PROP_MASS	=   13900.0; //*** the total mass of fueled CentaurD is 15600kg
const double ISP_CHARIOT_SL		=     365.0*G; 
const double ISP_CHARIOT_VAC	=     422.0*G; //***
const double THRUST_CHARIOT_VAC	=   15590.0*G; //***
const double CHARIOT_PROP_MASS	=   14200.0; //*** the total mass of fueled CHARIOT is 16000kg

const double R_EARTH = 6371000;			// radius of the earth (m)
const double U_EARTH = 3.986e5 * 1000 * 1000 * 1000; //( m^3/sec^2)
const double APO_ALT = 140.0;


const VECTOR3 X20_OFS			= _V(0,0,0);
const VECTOR3 BAY1_OFS			= _V(0,0,0);
const VECTOR3 BAY2_OFS			= _V(0,0,0);
const VECTOR3 TOPHATCH_OFS		= _V(0,0,0);
const VECTOR3 ENTRYSHIELD_OFS	= _V(0,0,0);
const VECTOR3 EJSEAT_OFS		= _V(0.0,-0.57,6.6);
const VECTOR3 ASTRO_OFS			= _V(0.0,-0.45,6.8);
const VECTOR3 TRANSITION_OFS	= _V(0,0,0);
const VECTOR3 TRANSTAGE_OFS		= _V(0,0,0);
const VECTOR3 CAMERA_OFS		= _V(0,-0.04,6.62);

const double GEAR_OPERATING_SPEED = 0.25;
const double PLDOORS_OPERATING_SPEED = 0.07125;
const double BRAKES_OPERATING_SPEED = 0.25;
const double HATCH_OPERATING_SPEED = 0.125;

#define PLAYGEAR					 20		// the gear sound
#define PLAYBANG					 24		// the jettison sound
#define PLAYJETT					 25     // entry shield jettison sound

/* make your own vessel class (named X20 here, for example)
   subclassed from COMMON_VESSEL_EX */

class X20 : public COMMON_VESSEL_EX {
public:
	X20 (OBJHANDLE hObj, int fmodel);
	~X20();

	enum DoorStatus { DOOR_CLOSED, DOOR_OPEN, DOOR_CLOSING, DOOR_OPENING } gear_status, pldoors_status, brakes_status, hatch_status;

	void UpdateAnimations();
	void ActivateLandingGear (DoorStatus action);
	void RevertLandingGear();
	void ActivatePayloadBayDoors (DoorStatus action);
	void RevertPayloadBayDoors();
	void ActivateBrakes (DoorStatus action);
	void RevertBrakes();
	void ActivateHatch (DoorStatus action);
	void RevertHatch();
	void SeparateEVA();
	void SetGearParameters (double state);
	void JettisonEntryShield();
	void AttachSweep();
	void TimeStep(  double simt ); // need a timestep loop to conduct the GPDS polling
	void VisualCreated( VESSEL2 *vessel, VISHANDLE vis, int refcount ); // must use VisualCreated() instead of ovcVisualCreated()
	void SetClassCaps(  FILEHANDLE cfg );
	void LoadState(  FILEHANDLE scn, void *vs );
	void SaveState(  FILEHANDLE scn );
	void PlaceComponents();

	/* CVE-lite requirement ! */
	// Create a function that can be called as many times as you like. It is
	// vital to have a function like this that cleans out all the
	// thrusters and meshes and draws/recreates them again.
	// This is so that you can respond immediately to GPDS requests.
	// Here I've called it SetMainStage() it can be Refresh() or anything you like.
	void SetMainStage(); 
	void SetTranstage();
	void SetCentaur();
	void SetChariot();
	void SetDeorbitStage();
	void SetX20Stage();

	void SetAttControlsTranstage();
	void SetAttControlsX20();    // this is to set up the thrusters. I like to keep them separate.
	void JettisonPart();
	int  ConsumeKey(  const char *keystate );
	int  ConsumeBufferedKey(DWORD dwKey, bool fPressed, const char *keystate);

// Overloaded callback functions
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkLoadStateEx (FILEHANDLE scn, void *vs);
	void clbkSaveState (FILEHANDLE scn);
	void clbkPostStep (double simt, double simdt, double mjd);
	int  clbkConsumeBufferedKey (DWORD key, bool down, char *kstate);
	void clbkVisualCreated (VISHANDLE vis, int refcount);
	void clbkVisualDestroyed (VISHANDLE vis, int refcount);
	bool clbkLoadVC (int id);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);

protected:
	PROPELLANT_HANDLE ph_first,ph_x20,ph_deorbit;
	THRUSTER_HANDLE   th_first[4];
	THGROUP_HANDLE    thg_first,thg_pitchup, thg_pitchdown;
	MESHHANDLE		  hX20,hX20Bay1,hX20Bay2,hTopHatch,hEntryShield,hSeat,hAstro,hTitanTransition,hTranstage,hCentaur,hChariot,hDummy1,hDummy2,hDummy3,hDummy4, hPit;
	SURFHANDLE		  tex,texRCS, retex;
	int stage;
	UINT anim_lrudder,anim_rrudder,anim_laileron,anim_raileron,anim_lelevator,anim_relevator,anim_lelevatortrim,anim_relevatortrim;
	UINT anim_gear;     // handle for landing gear animation
	UINT anim_pldoors, anim_pldoors2;     // handle for payload door animation
	UINT anim_brakes;	// handle for speedbrake animation
	UINT anim_hatch;	// handle for top hatch animation
	UINT anim_pitch, anim_roll, anim_yaw;
	double gear_proc;
	double pldoors_proc;
	double brakes_proc;
	double hatch_proc;
	double pitch_proc, roll_proc, yaw_proc;
	ANIMATIONCOMPONENT_HANDLE animh_yaw, animh_roll, animh_pitch;
	bool BrakesOn, havesound, bShieldJettisoned, bDeorbitFired, bThrustTerminated, bEjected, bEVA, bAbort, AbortChecked;
	UINT Exh1,Exh2,Exh3,Exh4;
	PSTREAM_HANDLE stream_a_1,stream_b_1,stream_a_2,stream_b_2,stream_a_3,stream_b_3,stream_a_4,stream_b_4;
	VECTOR3 Vold, Vnew;
	double dt, oldsimt;
	int MyID;
	ATTACHMENTHANDLE Baypos;
	VECTOR3 gpos, grms, pos, dir, rot;
	OBJHANDLE hV, BayVessel;
	VESSEL *BayV;
	DWORD i, j;
	double atchmass;
	double a,e,h,V_periapsis,Altitude;
	DOCKHANDLE hDock, Dock0;
	bool Firstsweep, AutoPilot;
	char *key;
	char *message;
	VECTOR3 Rel_Vel;
	double Velocity;
	int stagetype;
	double Pitch_SP, Pitch_Error, Pitch_Error_Dot, Pitch_Cntrl_Sig, Roll_Cntrl_Sig, Alt_Apoapsis, Alt_Periapsis, MJD_Time;
	VECTOR3 Angular_Velocity, horiz;
	ELEMENTS elements;
	VCHUDSPEC huds;
	VCMFDSPEC mfds_left;
	VCMFDSPEC mfds_right;
	VECTOR3 yawvec[1];
	VECTOR3 pitchvec[1];


};




