/**
*
* $Id: PAXMOD.cpp,v 0.1 2004/11/20 18:03:45 erv Exp $
* (c) 2004 Sputnik -- erik_h_anderson@hotmail.com
* Based on CVEL Centaur-G code
* Ervins Reinverts -- ervins@gmail.com
* Based on CVEL Fregat code
* (c) 2004 David Henderson -- orbiter@aibs.org.uk
* Licence LGPL -- http://www.gnu.org/licenses/lgpl.txt
*
* You may use this source code in your addon provided that if you
* publish your addon either privately or publicly the end user
* recieves a copy of the source code including any modifications
* you may have made. 
*
*/

#include "PAXMOD.h"



PAXMOD::PAXMOD (OBJHANDLE hObj, int fmodel)
: COMMON_VESSEL_EX (hObj)
{
	hPAXMOD				= oapiLoadMeshGlobal ("X20/Paxbay");
	pldoors_status=DOOR_CLOSED;
	pldoors_proc=0.0;
	anim_pldoors = 0.0;

	/* CVE-lite function ! */
	/* IMPORTANT */
	/* This function loops through the pending GDPS requests. Have it run
	/* in the constructor so it can set up things like the InitFuelLevel
	/* (initial fuel level) variable, and load the payloads */
	CheckCentralMessage();   

	static UINT DockPortGrp[1] = {0};
	static MGROUP_TRANSLATE DockPort (0, DockPortGrp, 1,
		_V(0.0,0.2,0.0));

	anim_pldoors = CreateAnimation(0);
	AddAnimationComponent (anim_pldoors, 0.1, 0.9, &DockPort);
	RevertPayloadBayDoors();


	key = (char*)VirtualAlloc(NULL,256,MEM_COMMIT,PAGE_READWRITE);
	message = (char*)VirtualAlloc(NULL,256,MEM_COMMIT,PAGE_READWRITE);
}


PAXMOD::~PAXMOD()
{

//	delete []key;
//	delete []message;
	VirtualFree(key,256,MEM_DECOMMIT);
	VirtualFree(message,256,MEM_DECOMMIT);
}

void PAXMOD::UpdateAnimations()
{	

	// animate payload bay doors
	if (pldoors_status >= DOOR_CLOSING) {
		double da = oapiGetSimStep() * PLDOORS_OPERATING_SPEED;
		if (pldoors_status == DOOR_CLOSING) {
			if (pldoors_proc > 0.0)
				pldoors_proc = max (0.0, pldoors_proc-da);
			else {
				pldoors_status = DOOR_CLOSED;
			}
		} else  { // door opening
			if (pldoors_proc < 1.0)
				pldoors_proc = min (1.0, pldoors_proc+da);
			else {
				pldoors_status = DOOR_OPEN;
			}
		}
		SetAnimation (anim_pldoors, pldoors_proc);
	}

}

void PAXMOD::ActivatePayloadBayDoors (DoorStatus action)
{
	pldoors_status = action;
}

void PAXMOD::RevertPayloadBayDoors ()
{
	ActivatePayloadBayDoors (pldoors_status == DOOR_CLOSED || pldoors_status == DOOR_CLOSING ?
						 DOOR_OPENING : DOOR_CLOSING);
}


void PAXMOD::TimeStep(  double simt )
{
	
	/* CVE-lite function ! */
	/* IMPORTANT */
	/* This function loops through the pending GPDS requests and returns a bool (true/false)
	/* if any GPDS requests have been implemented.
	/* If any have (eg adding a new payload), you will need to re-draw your ship, hence the
	/* call to the (repeatable) SetMainStage(); */

	ZeroMemory (key,256);
	ZeroMemory (message,256);

	if( CheckCentralMessage(key, message) ) {
		CheckCentralMessage();
		if (( strstr(key, "MSG" ) ) &&( strstr(message, "REVERTPROTRUSIONS" ) )) RevertPayloadBayDoors();
		while( CheckCentralMessage() ) ;
			SetMainStage();
	}


	UpdateAnimations();

}

/* CVE-lite function ! */
/* IMPORTANT */
/* VisualCreated() is a function passed through from CVE-lite as a direct replacement
/* for the DLLCLBK ovcVisualCreated() function.
/* You can't use ovcVisualCreated() with CVE-lite, so use this instead! */

void PAXMOD::VisualCreated ( VESSEL *vessel, VISHANDLE vis, int refcount) {
	/* local equivalent to the DLLCLBK ovcVisualCreated() */
	SetMainStage(); // this is a good place to ask for the initial setup.
}


void PAXMOD::SetClassCaps(  FILEHANDLE cfg )
{

	CheckCentralMessage();

	SetMainStage();
}

void PAXMOD::SaveState( FILEHANDLE scn )
{
	char cbuf[256];

	SaveDefaultState (scn);
	
	/* CVE-lite function ! */
	/* IMPORTANT */
	/* Call this after the normal SaveDefautState() to record attached payloads
	/* it the scenario file */

	sprintf (cbuf, "%d %0.4f", pldoors_status, pldoors_proc);
	oapiWriteScenario_string (scn, "DOORS", cbuf);

	/* save payload lines */
	SaveDefaultStateEx(scn);
}

void PAXMOD::LoadState(  FILEHANDLE scn, void *vs )
{
    char *line;

	while (oapiReadScenario_nextline (scn, line)) 
	{
		/* CVE-lite function ! */
		/* IMPORTANT */
		/* Call this for every line in the scenario file. It will automatically
		/* deduce which lines are for it, and load the payloads from the file. */

		/* load payload lines */
		if (!strnicmp (line, "DOORS", 5)) {
		sscanf (line+5, "%d%lf", &pldoors_status, &pldoors_proc);
		}
		ParseScenarioLineEx (line, vs);
	}

	/* always set the stage whenever something may have changed */
	SetMainStage();
	RevertPayloadBayDoors();
	RevertPayloadBayDoors();
//	if ((pldoors_status == DOOR_OPEN) && (pldoors_proc == 1.0)) RevertPayloadBayDoors();

}


void PAXMOD::PlaceComponents()
{
	COMMON_MESH tmp;


	tmp.MeshHandle(hPAXMOD);
	tmp.MeshPosition(_V(0,0,0));
	AddMeshEx (tmp);

}

/* IMPORTANT */
/* Always have a function that you can call as many times as you like that will re-draw
/* all you meshes, reset all your thrusters and check to make sure the propellent is
/* there. This should be called after any configuration change, or any GDPS line (see
/* TimeStep(), above) */

void PAXMOD::SetMainStage( )
{

	ClearMeshesEx();
	PlaceComponents();
	Draw();

}


int PAXMOD::ConsumeBufferedKey(DWORD dwKey, bool fPressed, const char *keystate)
{
	int result = 0;

	if(!fPressed && !KEYMOD_SHIFT(keystate) && !KEYMOD_CONTROL(keystate) && !KEYMOD_ALT(keystate))
	{
		switch(dwKey)
		{ 

		case OAPI_KEY_K:
				// Toggle the bay doors
				RevertPayloadBayDoors ();
//				if (BayVessel != NULL)
//					SendCentralMessage(BayV, "MSG", "REVERTPROTRUSIONS");
//				if (pldoors_status == DOOR_CLOSING)
//					Undock(ALLDOCKS);
			result = 1;
			break;		
		}
	}
	return result;
}



int PAXMOD::ConsumeKey( const char *keystate )
{

	if (KEYMOD_SHIFT (keystate)) 
	{
						
		return 0; // shift combinations are reserved
	}
	else if (KEYMOD_CONTROL (keystate)) 
	{
		return 0;
	}
	else 
	{ // unmodified keys
	
		return 0;

	}
	return 0;
}



